// $Id: EOQualifier.m,v 1.7 2003/02/03 15:58:02 helge Exp $

#include <stdio.h>
#include "EOQualifier.h"
#include "EOKeyValueCoding.h"
#include "common.h"
#include "EONull.h"
#import <Foundation/NSException.h>

@interface NSObject(QualifierDescription)
- (NSString *)qualifierDescription;
@end

@implementation EOQualifier

static NSMapTable *operatorToSelector = NULL;
static NSMapTable *selectorToOperator = NULL;
static EONull     *null = nil;

+ (void)initialize {
  if (null == nil) null = [EONull null];
  
  if (operatorToSelector == NULL) {
    operatorToSelector = NSCreateMapTable(NSObjectMapKeyCallBacks,
                                          NSIntMapValueCallBacks,
                                          10);
    NSMapInsert(operatorToSelector, @"=",   EOQualifierOperatorEqual);
    NSMapInsert(operatorToSelector, @"==",  EOQualifierOperatorEqual);
    NSMapInsert(operatorToSelector, @"!=",  EOQualifierOperatorNotEqual);
    NSMapInsert(operatorToSelector, @"<>",  EOQualifierOperatorNotEqual);        
    NSMapInsert(operatorToSelector, @"<",   EOQualifierOperatorLessThan);
    NSMapInsert(operatorToSelector, @">",   EOQualifierOperatorGreaterThan);
    NSMapInsert(operatorToSelector, @"<=",  EOQualifierOperatorLessThanOrEqualTo);
    NSMapInsert(operatorToSelector, @"like",EOQualifierOperatorLike);
    NSMapInsert(operatorToSelector, @"LIKE",EOQualifierOperatorLike);
    NSMapInsert(operatorToSelector, @">=",
                EOQualifierOperatorGreaterThanOrEqualTo);
    NSMapInsert(operatorToSelector, @"caseInsensitiveLike",
                EOQualifierOperatorCaseInsensitiveLike);
  }
  if (selectorToOperator == NULL) {
    selectorToOperator = NSCreateMapTable(NSObjectMapKeyCallBacks,
                                          NSObjectMapValueCallBacks,
                                          10);
    NSMapInsert(selectorToOperator,
                NSStringFromSelector(EOQualifierOperatorEqual),
                @"=");
    NSMapInsert(selectorToOperator,
                NSStringFromSelector(EOQualifierOperatorNotEqual),
                @"<>");
    NSMapInsert(selectorToOperator,
                NSStringFromSelector(EOQualifierOperatorLessThan),
                @"<");
    NSMapInsert(selectorToOperator,
                NSStringFromSelector(EOQualifierOperatorGreaterThan),
                @">");
    NSMapInsert(selectorToOperator,
                NSStringFromSelector(EOQualifierOperatorLessThanOrEqualTo),
                @"<=");
    NSMapInsert(selectorToOperator,
                NSStringFromSelector(EOQualifierOperatorLike),
                @"like");
    NSMapInsert(selectorToOperator,
                NSStringFromSelector(EOQualifierOperatorGreaterThanOrEqualTo),
                @">=");
    NSMapInsert(selectorToOperator,
                NSStringFromSelector(EOQualifierOperatorCaseInsensitiveLike),
                @"caseInsensitiveLike");
  }
}

+ (EOQualifier *)qualifierToMatchAnyValue:(NSDictionary *)_values {
  /* OR qualifier */
  NSEnumerator *keys;
  NSString     *key;
  NSArray      *array;
  unsigned i;
  id qualifiers[[_values count] + 1];
  
  keys = [_values keyEnumerator];
  for (i = 0; (key = [keys nextObject]); i++) {
    id value;
    
    value = [_values objectForKey:key];
    qualifiers[i] =
      [[EOKeyValueQualifier alloc]
                            initWithKey:key
                            operatorSelector:EOQualifierOperatorEqual
                            value:value];
    qualifiers[i] = [qualifiers[i] autorelease];
  }
  array = [NSArray arrayWithObjects:qualifiers count:i];
  return [[[EOOrQualifier alloc] initWithQualifierArray:array] autorelease];
}

+ (EOQualifier *)qualifierToMatchAllValues:(NSDictionary *)_values {
  /* AND qualifier */
  NSEnumerator *keys;
  NSString     *key;
  NSArray      *array;
  unsigned i;
  id qualifiers[[_values count] + 1];
  
  keys = [_values keyEnumerator];
  for (i = 0; (key = [keys nextObject]); i++) {
    id value;
    
    value = [_values objectForKey:key];
    qualifiers[i] =
      [[EOKeyValueQualifier alloc]
                            initWithKey:key
                            operatorSelector:EOQualifierOperatorEqual
                            value:value];
    qualifiers[i] = [qualifiers[i] autorelease];
  }
  array = [NSArray arrayWithObjects:qualifiers count:i];
  return [[[EOAndQualifier alloc] initWithQualifierArray:array] autorelease];
}

+ (SEL)operatorSelectorForString:(NSString *)_str {
  SEL s;

  if ((s = NSMapGet(operatorToSelector, _str)))
    return s;
  else
    return NSSelectorFromString(_str);
}

+ (NSString *)stringForOperatorSelector:(SEL)_sel {
  NSString *s, *ss;
  
  if ((s = NSStringFromSelector(_sel)) == nil)
    return nil;
  
  if ((ss = NSMapGet(selectorToOperator, s)))
    return ss;
  
  return s;
}

/* inits */

- (id)initWithDictionary:(NSDictionary *)_dict {
  [self release];
  return [[EOQualifier qualifierToMatchAllValues:_dict] retain];
}

- (id)initWithArray:(NSArray *)_array {
  unsigned count;
  NSString *fmt;
  NSArray  *args;
  
  [self release];

  if ((count = [_array count]) == 0) {
    NSLog(@"%s: invalid array for qualifier: %@", __PRETTY_FUNCTION__, _array);
    return nil;
  }
  
  fmt = [_array objectAtIndex:0];
  if (count == 1)
    args = nil;
  else
    args = [_array subarrayWithRange:NSMakeRange(1, (count - 1))];
  
  return [[EOQualifier qualifierWithQualifierFormat:fmt arguments:args]
                       retain];
}

- (id)initWithString:(NSString *)_string {
  [self release];
  return [[EOQualifier qualifierWithQualifierFormat:_string] retain];
}

- (id)initWithPropertyList:(id)_plist {
  if ([_plist isKindOfClass:[NSDictionary class]])
    return [self initWithDictionary:_plist];
  else if ([_plist isKindOfClass:[NSString class]])
    return [self initWithString:_plist];
  else if ([_plist isKindOfClass:[NSArray class]])
    return [self initWithArray:_plist];
  else if ([_plist isKindOfClass:[self class]]) {
    [self release];
    return [_plist copy];
  }
  else {
    [self release];
    return nil;
  }
}

/* bindings */

- (EOQualifier *)qualifierWithBindings:(NSDictionary *)_bindings
  requiresAllVariables:(BOOL)_reqAll
{
  return self;
}

- (NSArray *)bindingKeys {
  return nil;
}

/* keys */

- (NSSet *)allQualifierKeys {
  /* new in WO 4.5 */
  id set;

  set = [NSMutableSet setWithCapacity:64];
  [self addQualifierKeysToSet:set];
  return [[set copy] autorelease];
}

- (void)addQualifierKeysToSet:(NSMutableSet *)_keys {
  /* new in WO 4.5 */
}

/* Comparing */

- (BOOL)isEqual:(id)_obj {
  if ([_obj isKindOfClass:[self class]] == YES) {
    return [self isEqualToQualifier:(EOQualifier *)_obj];
  }
  return NO;
}

- (BOOL)isEqualToQualifier:(EOQualifier *)_qual {
  [self doesNotRecognizeSelector:_cmd];
  return NO;
}

/* remapping keys */

- (EOQualifier *)qualifierByApplyingTransformer:(id)_t inContext:(id)_ctx {
  if ([_t respondsToSelector:@selector(transformQualifier:inContext:)])
    return [_t transformQualifier:self inContext:_ctx];
  else
    return [[self retain] autorelease];
}
- (EOQualifier *)qualifierByApplyingKeyMap:(NSDictionary *)_key {
  return [[self copy] autorelease];
}

/* debugging */

+ (BOOL)isEvaluationDebuggingEnabled {
  static int evalDebug = -1;
  if (evalDebug == -1) {
    evalDebug = [[NSUserDefaults standardUserDefaults] 
                  boolForKey:@"EOQualifierDebugEvaluation"] ? 1 : 0;
    if (evalDebug)
      NSLog(@"WARNING: qualifier evaluation debugging is enabled !");
  }
  return evalDebug ? YES : NO;
}

@end /* EOQualifier */

@implementation EOQualifier(QuickEval)

- (BOOL)evaluateWithObject:(id)_object inEvalContext:(id)_ctx {
  return [(id<EOQualifierEvaluation>)self evaluateWithObject:_object];
}

@end /* EOQualifier(QuickEval) */

/* variable qualifier content */

@implementation EOQualifierVariable

+ (id)variableWithKey:(NSString *)_key {
  return [[[self alloc] initWithKey:_key] autorelease];
}

- (id)initWithKey:(NSString *)_key {
  self->varKey = [_key copyWithZone:[self zone]];
  return self;
}
- (id)init {
  return [self initWithKey:nil];
}

- (void)dealloc {
  [self->varKey release];
  [super dealloc];
}

/* accessors */

- (NSString *)key {
  return self->varKey;
}

/* NSCoding */

- (void)encodeWithCoder:(NSCoder *)_coder {
  [_coder encodeObject:self->varKey];
}
- (id)initWithCoder:(NSCoder *)_coder {
  self->varKey = [[_coder decodeObject] copyWithZone:[self zone]];
  return self;
}

/* Comparing */

- (BOOL)isEqual:(id)_obj {
  if ([_obj isKindOfClass:[self class]] == YES) {
    return [self isEqualToQualifierVariable:(EOQualifierVariable *)_obj];
  }
  return NO;
}

- (BOOL)isEqualToQualifierVariable:(EOQualifierVariable *)_obj {
  return [self->varKey isEqual:[_obj key]];
}

/* description */

- (NSString *)qualifierDescription {
  return [@"$" stringByAppendingString:[self key]];
}

- (NSString *)description {
  return [NSString stringWithFormat:@"<%@[0x%08X]: variable=%@>",
                     NSStringFromClass([self class]), self,
                     [self key]];
}

@end /* EOQualifierVariable */

@implementation NSObject(QualifierDescription)

- (NSString *)qualifierDescription {
  return [self description];
}

@end /* NSObject(QualifierDescription) */

@implementation NSString(QualifierDescription)

- (NSString *)qualifierDescription {
  return [NSString stringWithFormat:@"'%@'", self];
}

@end /* NSString(QualifierDescription) */

@implementation EONull(QualifierDescription)

- (NSString *)qualifierDescription {
  return @"nil";
}

@end /* EONull(QualifierDescription) */
