// $Id: EOSortOrdering.m,v 1.5 2003/04/01 13:32:30 gnustep Exp $

#include "EOSortOrdering.h"
#include "EOKeyValueCoding.h"
#include <EOControl/EONull.h>
#include "common.h"

#if !NeXT_RUNTIME
EOControl_DECLARE SEL EOCompareAscending                 = 
  @selector(compareAscending:);
EOControl_DECLARE SEL EOCompareDescending                = 
  @selector(compareDescending:);
EOControl_DECLARE SEL EOCompareCaseInsensitiveAscending  =
  @selector(compareCaseInsensitiveAscending:);
EOControl_DECLARE SEL EOCompareCaseInsensitiveDescending = 
  @selector(compareCaseInsensitiveDescending:);
#endif

@implementation EOSortOrdering
/*"
  This class specifies a sort-ordering as used with
  EOFetchSpecification. It takes a key and a sort
  selector which is used for comparision.
"*/

/*" Create a sort-ordering object with the specified key and sort selector "*/
+ (EOSortOrdering *)sortOrderingWithKey:(NSString *)_key selector:(SEL)_sel {
  return [[[self alloc] initWithKey:_key selector:_sel] autorelease];
}

/*" 
  Initialize a sort-ordering object with the specified key and sort selector
"*/
- (id)initWithKey:(NSString *)_key selector:(SEL)_selector {
  if ((self = [super init])) {
    self->key      = [_key copyWithZone:[self zone]];
    self->selector = _selector;
  }
  return self;
}

/*"
  Initialize a sort-ordering with information contained in the dictionary.
  The following keys are recognized: "key" is required and specifies the
  key to be sorted on, "selector" is optional and specifies the sort
  selector as a string. The default for "selector" is EOCompareAscending
  and the following "special" values are recognized: "compareAscending",
  "compareDescending", "compareCaseInsensitiveAscending", 
  "compareCaseInsensitiveDescending".
"*/
- (id)initWithDictionary:(NSDictionary *)_dict {
  NSString *k  = nil;
  SEL      sel = EOCompareAscending;
  NSString *tmp;

  if (_dict == nil) {
    [self release];
    return nil;
  }
  
  k = [_dict objectForKey:@"key"];
  if ([k length] == 0) {
    NSLog(@"%s: invalid key %@ (dict=%@)", __PRETTY_FUNCTION__, k, _dict);
    [self release];
    return nil;
  }
  
  if ((tmp = [[_dict objectForKey:@"selector"] stringValue])) {
    if ([tmp isEqualToString:@"compareAscending"])
      sel = EOCompareAscending;
    else if ([tmp isEqualToString:@"compareDescending"])
      sel = EOCompareDescending;
    else if ([tmp isEqualToString:@"compareCaseInsensitiveAscending"])
      sel = EOCompareDescending;
    else if ([tmp isEqualToString:@"compareCaseInsensitiveDescending"])
      sel = EOCompareDescending;
    else
      sel = NSSelectorFromString(tmp);
  }
  return [self initWithKey:key selector:sel];
}

/*"
  Initialize/parse a sort-ordering from a string. Usually the string is
  taken as the key of the ordering and the sorting EOCompareAscending. This
  can be modified by adding ".reverse" to the key, eg "name.reverse" sorts
  on the "name" key using EOCompareDescending.
"*/
- (id)initWithString:(NSString *)_string {
  SEL      sel;
  NSString *k;
  NSRange  r;
  
  if ([_string length] == 0) {
    [self release];
    return nil;
  }
  
  r = [_string rangeOfString:@".reverse"];
  if (r.length == 0) {
    k    = _string;
    sel = EOCompareAscending;
  }
  else {
    k   = [_string substringToIndex:r.location];
    sel = EOCompareDescending;
  }
  
  return [self initWithKey:k selector:sel];
}

/*"
  This method checks whether the give property list is an NSDictionary
  or an NSString object and invokes -initWithDictionary: or 
  -initWithString: in turn.
"*/
- (id)initWithPropertyList:(id)_plist {
  if (_plist == nil) {
    [self release];
    return nil;
  }
  
  if ([_plist isKindOfClass:[NSDictionary class]])
    return [self initWithDictionary:_plist];
  if ([_plist isKindOfClass:[NSString class]])
    return [self initWithString:_plist];
  
  if ([_plist isKindOfClass:[self class]]) {
    [self release];
    return [_plist copy];
  }
  
  [self release];
  return nil;
}

- (void)dealloc {
  [self->key release];
  [super dealloc];
}

/* accessors */

/*"
  Returns the key the ordering sorts with.
"*/
- (NSString *)key {
  return self->key;
}

/*"
  Returns the selector the ordering sorts with.
"*/
- (SEL)selector {
  return self->selector;
}

/* equality */

- (BOOL)isEqualToSortOrdering:(EOSortOrdering *)_sortOrdering {
  if (!SEL_EQ([_sortOrdering selector], [self selector]))
    return NO;
  if (![[_sortOrdering key] isEqualToString:[self key]])
    return NO;
  return YES;
}
- (BOOL)isEqual:(id)_other {
  if ([_other isKindOfClass:[EOSortOrdering class]])
    return [self isEqualToSortOrdering:_other];

  return NO;
}

/* remapping keys */

- (EOSortOrdering *)sortOrderingByApplyingKeyMap:(NSDictionary *)_map {
  NSString *k;
  
  k = [_map objectForKey:self->key];
  return [EOSortOrdering sortOrderingWithKey:(k ? k : self->key)
			 selector:self->selector];
}

/* description */

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<0x%08X[%@]: key=%@ selector=%@>",
                     self, NSStringFromClass([self class]),
                     [self key], NSStringFromSelector([self selector])];
}

@end /* EOSortOrdering */

@implementation NSArray(EOSortOrdering)

/*" 
  Sort the array using the sort-orderings contained in the argument. If no
  orderings are given, a copy of self is returned.
"*/
- (NSArray *)sortedArrayUsingKeyOrderArray:(NSArray *)_orderings {
  NSMutableArray *m      = nil;
  NSArray        *result = nil;
  
  if ([_orderings count] == 0)
    return [[self copy] autorelease];
  
  m = [self mutableCopy];
  [m sortUsingKeyOrderArray:_orderings];
  result = [m copy];
  [m release]; m = nil;
  return [result autorelease];
}

@end /* NSArray(EOSortOrdering) */

@implementation NSMutableArray(EOSortOrdering)

typedef struct {
  EOSortOrdering *orderings[10]; /* max depth 10 */
  short          count;
} EOSortOrderingContext;

static EONull *null = nil;

static int keyOrderComparator(id o1, id o2, EOSortOrderingContext *context) {
  short i;

  for (i = 0; i < context->count; i++) {
    NSString *key;
    SEL      sel;
    id       v1, v2;
    int      (*ccmp)(id, SEL, id);
    int      result;

    key = [context->orderings[i] key];
    sel = [context->orderings[i] selector];
    
    v1 = [o1 valueForKey:key];
    v2 = [o2 valueForKey:key];

    if (v1 == v2)
      result = NSOrderedSame;
    else if ((v1 == nil) || (v1 == null))
      result = (sel == EOCompareAscending)
        ? NSOrderedAscending : NSOrderedDescending;
    else if ((v2 == nil) || (v2 == null))
      result = (sel == EOCompareAscending)
        ? NSOrderedDescending : NSOrderedAscending;
    else if ((ccmp = (void *)[v1 methodForSelector:sel]))
      result = ccmp(v1, sel, v2);
    else
      result = (int)[v1 performSelector:sel withObject:v2];

    if (result != NSOrderedSame)
      return result;
  }
  return NSOrderedSame;
}

/*" 
  Sort the array using the sort-orderings contained in the argument.
"*/
- (void)sortUsingKeyOrderArray:(NSArray *)_orderings {
  NSEnumerator          *e        = nil;
  EOSortOrdering        *ordering = nil;
  EOSortOrderingContext ctx;
  int                   i;
  
  e = [_orderings objectEnumerator];
  for (i = 0; (ordering = [e nextObject]) && (i < 10); i++)
    ctx.orderings[i] = ordering;

  ctx.count = i;

  if (null == nil) null = [EONull null];
  [self sortUsingFunction:(void *)keyOrderComparator context:&ctx];
}

@end /* NSMutableArray(EOSortOrdering) */

@implementation EONull(EOSortOrdering)

/*" 
  Compares the null object, "nil" and "self" are considered of the same order,
  otherwise null is considered of lower order.
"*/
- (int)compareAscending:(id)_object {
  if (_object == self) return NSOrderedSame;
  if (_object == nil)  return NSOrderedSame;
  return NSOrderedDescending;
}

/*" 
  Compares the null object, "nil" and "self" are considered of the same order,
  otherwise null is considered of higher order.
"*/
- (int)compareDescending:(id)_object {
  if (_object == self) return NSOrderedSame;
  if (_object == nil)  return NSOrderedSame;
  return NSOrderedAscending;
}

@end /* EONull(EOSortOrdering) */

@implementation NSNumber(EOSortOrdering)

static Class NumClass = Nil;

- (int)compareAscending:(id)_object {
  if (_object == self) return NSOrderedSame;
  if (NumClass == Nil) NumClass = [NSNumber class];
  
  if ([_object isKindOfClass:NumClass])
    return [self compare:_object];
  else
    return [_object compareDescending:self];
}

- (int)compareDescending:(id)_object {
  int result;
  
  result = [self compareAscending:_object];
  
  if (result == NSOrderedAscending)
    return NSOrderedDescending;
  else if (result == NSOrderedDescending)
    return NSOrderedAscending;
  else
    return NSOrderedSame;
}

@end /* NSNumber(EOSortOrdering) */

@implementation NSString(EOSortOrdering)

- (int)compareAscending:(id)_object {
  if (_object == self) return NSOrderedSame;
  return [self compare:[_object stringValue]];
}
- (int)compareCaseInsensitiveAscending:(id)_object {
  if (_object == self) return NSOrderedSame;
  return [self caseInsensitiveCompare:[_object stringValue]];
}

- (int)compareDescending:(id)_object {
  int result;
  
  if (_object == self) return NSOrderedSame;
  
  result = [self compareAscending:_object];
  
  if (result == NSOrderedAscending)
    return NSOrderedDescending;
  else if (result == NSOrderedDescending)
    return NSOrderedAscending;
  else
    return NSOrderedSame;
}

- (int)compareCaseInsensitiveDescending:(id)_object {
  int result;
  
  if (_object == self) return NSOrderedSame;
  result = [self compareCaseInsensitiveAscending:_object];
  
  if (result == NSOrderedAscending)
    return NSOrderedDescending;
  else if (result == NSOrderedDescending)
    return NSOrderedAscending;
  else
    return NSOrderedSame;
}

@end /* NSString(EOSortOrdering) */

@implementation NSDate(EOSortOrdering)

static Class DateClass = Nil;

- (int)compareAscending:(id)_object {
  if (_object == self) return NSOrderedSame;
  if (DateClass == Nil) DateClass = [NSDate class];
  if (![_object isKindOfClass:DateClass]) return NSOrderedAscending;
  return [self compare:_object];
}
- (int)compareDescending:(id)_object {
  int result;

  if (_object == self) return NSOrderedSame;

  if (DateClass == Nil) DateClass = [NSDate class];
  if (![_object isKindOfClass:DateClass]) return NSOrderedDescending;
  
  result = [self compare:_object];
  
  if (result == NSOrderedAscending)
    return NSOrderedDescending;
  else if (result == NSOrderedDescending)
    return NSOrderedAscending;
  else
    return NSOrderedSame;
}

@end /* NSDate(EOSortOrdering) */
