/* 
   NGCharBuffers.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGCharBuffers.h,v 1.2 2002/12/27 09:42:33 helge Exp $

#ifndef __NGExtensions_NGCharBuffers_H__
#define __NGExtensions_NGCharBuffers_H__

#include <string.h>
#include <ctype.h>
#import <Foundation/NSString.h>
#import <Foundation/NSException.h>
#import <NGExtensions/NGMemoryAllocation.h>

typedef struct {
  unsigned      capacity;
  unsigned      length;
  unsigned char increaseRatio;
  unsigned char *buffer; // zero terminated buffer
} NGCharBuffer8Struct;

typedef NGCharBuffer8Struct *NGCharBuffer8;

static inline NGCharBuffer8 NGCharBuffer8_init(NGCharBuffer8 _str,
                                               unsigned _capacity) {
  _str->capacity      = _capacity;
  _str->length        = 0;
  _str->increaseRatio = 2;
  _str->buffer        = NGMallocAtomic(_capacity);
  _str->buffer[0]     = '\0';
  return _str;
}
static inline void NGCharBuffer8_reset(NGCharBuffer8 _str) {
  if (_str) {
    _str->capacity = 0;
    _str->length   = 0;
    _str->increaseRatio = 0;
    NGFree(_str->buffer);
    _str->buffer = NULL;
  }
}

static inline NGCharBuffer8 NGCharBuffer8_new(unsigned _capacity) {
  NGCharBuffer8 str = NULL;
  str = NGMalloc(sizeof(NGCharBuffer8Struct));
  return NGCharBuffer8_init(str, _capacity);
}

static inline NGCharBuffer8 NGCharBuffer8_newWithCString(const char *_cstr) {
  NGCharBuffer8 str = NULL;
  str = NGMalloc(sizeof(NGCharBuffer8Struct));
  str = NGCharBuffer8_init(str, strlen(_cstr) + 2);
  strcpy(str->buffer, _cstr);
  return str;
}

static inline void NGCharBuffer8_dealloc(NGCharBuffer8 _str) {
  if (_str) {
    NGCharBuffer8_reset(_str);
    NGFree(_str);
    _str = NULL;
  }
}

static inline void NGCharBuffer8_checkCapacity(NGCharBuffer8 _str,
                                               unsigned _needed) {
  if (_str->capacity < (_str->length + _needed + 1)) {
    // increase size
    unsigned char *oldBuffer = _str->buffer;

    _str->capacity *= _str->increaseRatio;
    if (_str->capacity < (_str->length + _needed + 1))
      _str->capacity += _needed + 1;
    
    _str->buffer   = NGMallocAtomic(_str->capacity + 2);
    memcpy(_str->buffer, oldBuffer, (_str->length + 1));
    NGFree(oldBuffer);
    oldBuffer = NULL;
  }
}

static inline void NGCharBuffer8_addChar(NGCharBuffer8 _str, unsigned char _c) {
  NGCharBuffer8_checkCapacity(_str, 1);

  _str->buffer[_str->length] = _c;
  (_str->length)++;
  _str->buffer[_str->length] = '\0';
}

static inline void NGCharBuffer8_addCString(NGCharBuffer8 _str, char *_cstr) {
  unsigned len;
  
  if (_cstr == NULL)
    return;

  len = strlen(_cstr);
  NGCharBuffer8_checkCapacity(_str, len);
  strcat(_str->buffer, _cstr);
  _str->length += len;
}

static inline void NGCharBuffer8_removeContents(NGCharBuffer8 _str) {
  _str->length = 0;
  _str->buffer[0] = '\0';
}

static inline NSString *NGCharBuffer8_makeStringAndDealloc(NGCharBuffer8 _str) {
  if (_str == NULL) {
    return nil;
  }
  else {
    NSString *str = [NSString stringWithCString:_str->buffer length:_str->length];

    NSCAssert3(strlen(_str->buffer) == _str->length,
               @"length of cstring(%s) and the buffer do not match (%i vs %i)",
               _str->buffer, strlen(_str->buffer), _str->length);

    NGCharBuffer8_dealloc(_str); _str = NULL;
    return str;
  }
}

static inline void NGCharBuffer8_stripTrailingSpaces(NGCharBuffer8 _str) {
  if (_str == NULL)
    return;
  else if (_str->length == 0)
    return;
  else {
    while (_str->length > 0) {
      unsigned char c = _str->buffer[_str->length - 1];

      if (isspace((int)c) || (c == '\n') || (c == '\r')) {
        (_str->length)--;
      }
      else {
        break;
      }
    }
    _str->buffer[_str->length] = '\0';
  }
}

#endif /* __NGExtensions_NGCharBuffers_H__ */
