/* 
   NSException+misc.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSException+misc.h,v 1.3 2003/01/07 11:33:35 helge Exp $

#ifndef __NGExtensions_NSException_misc_H__
#define __NGExtensions_NSException_misc_H__

#import <Foundation/NSException.h>
#import <Foundation/NSLock.h>

/*
  Miscellaneous method additions to NSException

  the additions make it easier using exceptions in the Java way. In OpenStep
  exceptions are identified by their name, subclasses are mainly used when
  additional information needs to be stored. This works well as long as you do
  not use exceptions very much (as in the Foundation Kit) - if you do you
  soon wish to have a 'hierachy' of exceptions. This can be modelled - like in
  Java - using a subclass for each different kind of exception and then using
  the type of the exception for it's identification. This is supported by
  libFoundation using the TRY and CATCH macros.

  The methods below always use the class name as the exception name if the name
  isn't specified explicitly.
*/

@interface NSException(NGMiscellaneous)

- (id)initWithReason:(NSString *)_reason;
- (id)initWithReason:(NSString *)_reason userInfo:(id)_userInfo;
- (id)initWithFormat:(NSString *)_format,...;

@end

/*
  The following macros are for use of locks together with exception handling.
  A synchronized block is properly 'unlocked' even if an exception occures.
  It is used this way:

    SYNCHRONIZED(MyObject) {
      THROW(MyException..);
    }
    END_SYNCHRONIZED;

  Where MyObject must be an object that conforms to the NSObject and NSLocking
  protocol.
  This is much different to

    [MyObject lock];
    {
      THROW(MyException..);
    }
    [MyObject unlock];

  which leaves the lock locked when an exception happens.
*/

#if defined(DEBUG_SYNCHRONIZED)

#define SYNCHRONIZED(__lock__) \
  { \
    id<NSObject,NSLocking> __syncLock__ = [__lock__ retain]; \
    [__syncLock__ lock]; \
    fprintf(stderr, "0x%08X locked in %s.\n", \
            (unsigned)__syncLock__, __PRETTY_FUNCTION__); \
    NS_DURING {

#define END_SYNCHRONIZED \
    } \
    NS_HANDLER { \
      fprintf(stderr, "0x%08X exceptional unlock in %s exception %s.\n", \
              (unsigned)__syncLock__, __PRETTY_FUNCTION__,\
              [[localException description] cString]); \
      [__syncLock__ unlock]; \
      [__syncLock__ release]; __syncLock__ = nil; \
      [localException raise]; \
    } \
    NS_ENDHANDLER; \
    fprintf(stderr, "0x%08X unlock in %s.\n", \
            (unsigned)__syncLock__, __PRETTY_FUNCTION__); \
    [__syncLock__ unlock]; \
    [__syncLock__ release];  __syncLock__ = nil; \
  }

#else

#define SYNCHRONIZED(__lock__) \
  { \
    id<NSObject,NSLocking> __syncLock__ = [__lock__ retain]; \
    [__syncLock__ lock]; \
    NS_DURING {

#define END_SYNCHRONIZED \
    } \
    NS_HANDLER { \
      [__syncLock__ unlock]; \
      [__syncLock__ release]; \
      [localException raise]; \
    } \
    NS_ENDHANDLER; \
    [__syncLock__ unlock]; \
    [__syncLock__ release]; \
  }

#endif

#endif /* __NGExtensions_NSException_misc_H__ */
