/*
   NSString+Formatting.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSString+Formatting.m,v 1.6 2003/04/01 13:38:20 gnustep Exp $

#import "common.h"
#import "NSString+Formatting.h"

#if !LIB_FOUNDATION_LIBRARY && !GNUSTEP_BASE_LIBRARY
#  import <FoundationExt/objc-api.h>
#  import <FoundationExt/objc-runtime.h>
#endif

// format
//   %|[justification]|[fieldwidth]|[precision]|formatChar
//   %[-]{digit}*[.{digit}*]conv

static const char *formatChars =
  "diouxX" // integers
  "feEgG"  // double
  "c"      // int arg -> uchar
  "s"      // string
  "@"      // object
  "$"      // string object
;
static const char *formatAttrs = "#0- +,";
static const char *formatConv  = "hlLqZ";

typedef enum {
  DEFAULT_TYPE = 0,
  SHORT_TYPE,
  LONG_TYPE,
  VERY_LONG_TYPE
} NGFormatType;

// implementation

static inline NSString *
wideString(const char *str, int width, char fillChar, BOOL isRightAligned)
{
  if (width > 0) {
    int len = strlen(str);

    if (len < width) {
      unsigned char *tmp = objc_atomic_malloc(width + 1);
      register int cnt;

      if (isRightAligned) { // right aligned
        for (cnt = 0; cnt < (width - len); cnt++)
          tmp[cnt] = fillChar;
        strncpy(tmp + (width - len), str, len);
        tmp[width] = '\0';
      }
      else { // left aligned
        for (cnt = len; cnt < width; cnt++)
          tmp[cnt] = fillChar;
        strncpy(tmp, str, len);
        tmp[cnt] = '\0';
      }

      return [[[NSString alloc]
                         initWithCStringNoCopy:tmp length:strlen(tmp)
                         freeWhenDone:YES] autorelease];
    }
  }
  return [NSString stringWithCString:str];
}

static inline NSString *xsSimpleFormatObject(char *fmt, id _value, NSString *_ds) {
  /*
    formats an object value in an simple format
    (this is only the '%{mods}{type}' part of a real
    format string !)
    The method to get the value of the object is
    determined by the {type} of the Format,
    eg: i becomes intValue, @ becomes description ..

    the decimal point in float-strings can be replaced by the _ds
    String. To do this, the result of the Format is scanned for
    '.' and _ds is expanded for this.
  */
  int  fmtLen  = strlen(fmt);
  char fmtChar = fmt[fmtLen - 1];

  NGFormatType typeMode = DEFAULT_TYPE;
  int  width      = -1;
  int  prec       = -1;
  BOOL alignRight = YES;
  char *pos = fmt + 1;
  char *tmp;

  while (index(formatAttrs, *pos)) {
    if (*pos == '-')
      alignRight = NO;
    pos++;
  }
  
  // width
  tmp = pos;
  while (isdigit((int)*pos)) pos++;
  if (tmp != pos) {
    char old = *pos;
    *pos = '\0';
    width = atoi(tmp);
    *pos = old;
  }

  // prec
  if (*pos == '.') {
    pos++;
    tmp = pos;
    while (isdigit((int)*pos)) pos++;
    if (tmp != pos) {
      char old = *pos;
      *pos = '\0';
      prec = atoi(tmp);
      *pos = old;
    }
  }

  // conversion
  if (index(formatConv, *pos)) {
    switch (*pos) {
      case 'h':
        typeMode = SHORT_TYPE;
        pos++;
        break;
      case 'l':
        typeMode = LONG_TYPE;
        pos++;
        if (*pos == 'l') { // long-long
          typeMode = VERY_LONG_TYPE;
          pos++;
        }
        break;
      case 'L':
      case 'q':
        typeMode = VERY_LONG_TYPE;
        pos++;
        break;
    }
  }

  if (index("diouxX", fmtChar)) {
    char buf[128];
    int  len = -1;

    switch (typeMode) {
      case SHORT_TYPE: {
        unsigned short int i = [(NSNumber *)_value unsignedShortValue];
        len = sprintf(buf, fmt, i);
        break;
      }
      case LONG_TYPE: {
        unsigned long int i = [(NSNumber *)_value unsignedLongValue];
        len = sprintf(buf, fmt, i);
        break;
      }
      case VERY_LONG_TYPE: {
        long long i = [(NSNumber *)_value unsignedLongLongValue];
        len = sprintf(buf, fmt, i);
        break;
      }
      default: {
        unsigned int i = [(NSNumber *)_value unsignedIntValue];
        len = sprintf(buf, fmt, i);
        break;
      }
    }

    return (len >= 0) ? [NSString stringWithCString:buf length:len] : nil;
  }
  else if (fmtChar == 'c') {
    char buf[64];
    int  i = [(NSNumber *)_value unsignedCharValue];
    int  len;

    len = sprintf(buf, fmt, (unsigned char)i);

    return (len > 0) ? [NSString stringWithCString:buf length:len] : nil;
  }
  else if (index("feEgG", fmtChar)) {
    char buf[256];
    int  len;

    if (typeMode == VERY_LONG_TYPE) {
      long double i = [(NSNumber *)_value doubleValue];
      len = sprintf(buf, fmt, i);
    }
    else {
      double i = [(NSNumber *)_value doubleValue];
      len = sprintf(buf, fmt, i);
    }

    if (len >= 0) {
      if (_ds == nil)
        return [NSString stringWithCString:buf length:len];
      else {
        NSMutableString *result =
          [NSMutableString stringWithCapacity:len + [_ds cStringLength]];
        unsigned int cnt;
        char         *ptr = buf;
        unsigned int bufCount = 0;

        for (cnt = 0; cnt < len; cnt++) {
          if (buf[cnt] == '.') {
            if (bufCount > 0)
              [result appendString:[NSString stringWithCString:ptr length:bufCount]];

            [result appendString:_ds];
            
            ptr = &(buf[cnt + 1]);
            bufCount = 0;
          }
          else {
            bufCount++;
          }
        }
        
        if (bufCount > 0)
          [result appendString:[NSString stringWithCString:ptr length:bufCount]];
        return result;
      }
    }
    else return nil;
  }
  else {
    switch (fmtChar) {
      case 's': {
	unsigned len;
	char *buffer = NULL;
	id result;

	if (_value == nil)
	  return wideString("<null>", width, ' ', alignRight);
	
	len    = [(NSString *)_value cStringLength];
	buffer = malloc(len + 10);
	[_value getCString:buffer];
	buffer[len] = '\0';
	
        result = wideString(buffer, width, ' ', alignRight);
	free(buffer);
	return result;
      }
      
      case '@': {
        id       obj   = _value;
        NSString *dstr = obj ? [obj description] : @"<nil>";
	char     *buffer = NULL;
	unsigned len;

	if (dstr == nil)
	  return wideString("<null>", width, ' ', alignRight);
	
	len    = [dstr cStringLength];
	buffer = malloc(len + 10);
	[dstr getCString:buffer];
	buffer[len] = '\0';
	
	dstr = wideString(buffer, width, ' ', alignRight);
	free(buffer);
	return dstr;
      }
      
      case '$': {
        id       obj   = _value;
        NSString *dstr = obj ? [obj stringValue] : @"<nil>";
        char     *cstr = (char *)[dstr cString];

        if (cstr == NULL) cstr = "<null>";

        return wideString(cstr, width, ' ', alignRight);
      }
        
      default:
        fprintf(stderr, "unknown printf format used: '%s'\n", fmt);
        break;
    }
  }
  return nil;
}

static inline NSString *handleFormat(char *fmt, int fmtLen, va_list *ap) {
  char fmtChar;
  char typeMode   = DEFAULT_TYPE;
  int  width      = -1;
  int  prec       = -1;
  BOOL alignRight = YES;
  char *pos = fmt + 1;
  char *tmp;

  if (fmtLen == 0)
    return @"";

  fmtChar = fmt[fmtLen - 1];
  
  while (index(formatAttrs, *pos)) {
    if (*pos == '-')
      alignRight = NO;
    pos++;
  }
  
  // width
  tmp = pos;
  while (isdigit((int)*pos)) pos++;
  if (tmp != pos) {
    char old = *pos;
    *pos = '\0';
    width = atoi(tmp);
    *pos = old;
  }

  // prec
  if (*pos == '-') {
    pos++;
    tmp = pos;
    while (isdigit((int)*pos)) pos++;
    if (tmp != pos) {
      char old = *pos;
      *pos = '\0';
      prec = atoi(tmp);
      *pos = old;
    }
  }

  // conversion
  if (index(formatConv, *pos)) {
    switch (*pos) {
      case 'h':
        typeMode = SHORT_TYPE;
        pos++;
        break;
      case 'l':
        typeMode = LONG_TYPE;
        pos++;
        if (*pos == 'l') { // long-long
          typeMode = VERY_LONG_TYPE;
          pos++;
        }
        break;
      case 'L':
      case 'q':
        typeMode = VERY_LONG_TYPE;
        pos++;
        break;
    }
  }

  //  printf("  width=%i prec=%i\n", width, prec); fflush(stdout);
    
  if (index("diouxX", fmtChar)) {
    char buf[128];
    int  len = -1;

    switch (typeMode) {
      case SHORT_TYPE: {
        unsigned short int i = va_arg(*ap, int);
        len = sprintf(buf, fmt, i);
        break;
      }
      case LONG_TYPE: {
        unsigned long int i = va_arg(*ap, unsigned long int);
        len = sprintf(buf, fmt, i);
        break;
      }
      case VERY_LONG_TYPE: {
        long long i = va_arg(*ap, long long);
        len = sprintf(buf, fmt, i);
        break;
      }
      default: {
        unsigned int i = va_arg(*ap, unsigned int);
        len = sprintf(buf, fmt, i);
        break;
      }
    }

    return (len >= 0) ? [NSString stringWithCString:buf length:len] : nil;
  }
  else if (fmtChar == 'c') {
    char buf[64];
    int  i = va_arg(*ap, int);
    //    int  i = va_arg(*ap, unsigned char);
    int  len;

    if (i != 0) {
      len = sprintf(buf, fmt, (unsigned char)i);

      // xsprintf("got format %s char %i made %s len %i\n",
      //         fmt, i, buf, len);
    
      return (len > 0)
        ? [NSString stringWithCString:buf length:len] : nil;
    }
    else
      return @"<'\\0'-char>";
  }
  else if (index("feEgG", fmtChar)) {
    char buf[256];
    int  len;

    if (typeMode == VERY_LONG_TYPE) {
      long double i = va_arg(*ap, long double);
      len = sprintf(buf, fmt, i);
    }
    else {
      double i = va_arg(*ap, double);
      len = sprintf(buf, fmt, i);
    }

    return (len >= 0) ? [NSString stringWithCString:buf length:len] : nil;
  }
  else {
    id result = nil;
    
    switch (fmtChar) {
      case 's':
      case '@':
      case '$': {
        char *cstr = NULL;
	BOOL owned = NO;

        if (fmtChar == 's') {
          cstr = va_arg(*ap, char *);
        }
        else {
          id obj = va_arg(*ap, id);
          if (obj == nil)
            cstr = "<nil>";
          else {
	    NSString *d;
	    
            if ((d = (fmtChar == '@') ?[obj description]:[obj stringValue])) {
	      unsigned len = [d cStringLength];

	      cstr = objc_malloc(len + 1);
	      [d getCString:cstr];
	      cstr[len] = '\0';
	      owned = YES;
	    }
          }
        }

        if (cstr == NULL) cstr = "<null>";

        result = wideString(cstr, width, ' ', alignRight);
	if (owned) objc_free(cstr);
        break;
      }

      default:
        fprintf(stderr, "unknown printf format used: '%s'\n", fmt);
        break;
    }
    return result;
  }
  return nil;
}

static inline NSString *_stringWithCFormat(const char *_format, va_list _ap) {
  const char *firstPercent = index(_format, '%');

  //fprintf(stderr, "OWN: format='%s'\n", _format);
  //fflush(stderr);

  // first check whether there are any '%' in the format ..
  if (firstPercent == NULL) {
    // no formatting contained in _format
    return [NSString stringWithCString:_format];
  }
  else {
    NSMutableString *result = [NSMutableString stringWithCapacity:256];

    if ((firstPercent - _format) > 0) {
      NSString *s;
      s = [NSString stringWithCString:_format length:(firstPercent - _format)];
      if (s) [result appendString:s];
      _format = firstPercent;
    }

    while (*_format != '\0') { // until end of format string
      if (*_format == '%') { // found formatting character
        _format++; // skip '%'

        if (*_format == '%') { // was a quoted '%'
          [result appendString:@"%"];
          _format++;
        }
        else { // check format
          char extFmt[16];
          char *pos  = extFmt;

          extFmt[0] = '%';
          pos++;
        
          while ((*_format != '\0') &&
                 (index(formatChars, *_format) == NULL)) {
            *pos = *_format;
            _format++;
            pos++;
          }
          *pos = *_format;
          _format++;
          pos++;
          *pos = '\0';

          // printf("handling ext format '%s'\n", extFmt); fflush(stdout);
        
          [result appendString:handleFormat(extFmt, strlen(extFmt), &_ap)];
        }
      }
      else { // normal char
        const char *start = _format; // remember start
        _format++; // skip found char

        // further increase format until '\0' or '%'
        while ((*_format != '\0') && (*_format != '%'))
          _format++;
        
        [result appendString:[NSString stringWithCString:start
                                       length:(_format - start)]];
      }
    }
    return result;
  }
}

@implementation NSString(XSFormatting)

+ (id)stringWithCFormat:(const char *)_format arguments:(va_list)_ap {
  return [self stringWithString:_stringWithCFormat(_format, _ap)];
}

+ (id)stringWithFormat:(NSString *)_format arguments:(va_list)_ap {
  unsigned len;
  char     *cfmt;
  id s;

  len = [_format cStringLength] + 1;
  cfmt = malloc(len + 1);
  [_format getCString:cfmt]; cfmt[len] = '\0';
  s = [self stringWithString:_stringWithCFormat(cfmt, _ap)];
  free(cfmt);
  return s;
}

+ (id)stringWithCFormat:(const char *)_format, ... {
  id      result = nil;
  va_list ap;
  
  va_start(ap, _format);
  result = [self stringWithString:_stringWithCFormat(_format, ap)];
  va_end(ap);
  return result;
}

+ (id)stringWithFormat:(NSString *)_format, ... {
  id       result = nil;
  unsigned len;
  char     *cfmt;
  va_list  ap;
  
  len = [_format cStringLength];
  cfmt = malloc(len + 1);
  [_format getCString:cfmt]; cfmt[len] = '\0';
  va_start(ap, _format);
  result = [self stringWithString:_stringWithCFormat(cfmt, ap)];
  va_end(ap);
  free(cfmt);
  return result;
}

- (id)initWithFormat:(NSString *)_format arguments:(va_list)_ap {
  unsigned len;
  char *cfmt;

  len = [_format cStringLength];
  cfmt = malloc(len + 1);
  [_format getCString:cfmt]; cfmt[len] = '\0';
  self = [self initWithString:_stringWithCFormat(cfmt, _ap)];
  free(cfmt);
  return self;
}

@end /* NSString(XSFormatting) */

@implementation NSMutableString(XSFormatting)

- (void)appendFormat:(NSString *)_format arguments:(va_list)_ap {
  unsigned len;
  char *cfmt;

  len = [_format cStringLength];
  cfmt = malloc(len + 1);
  
  [_format getCString:cfmt]; cfmt[len] = '\0';
  [self appendString:_stringWithCFormat(cfmt, _ap)];
  free(cfmt);
}
- (void)appendFormat:(NSString *)_format, ... {
  unsigned len;
  char     *cfmt;
  va_list  ap;
  
  len = [_format cStringLength];
  cfmt = malloc(len + 1);
  [_format getCString:cfmt]; cfmt[len] = '\0';
  va_start(ap, _format);
  [self appendString:_stringWithCFormat(cfmt, ap)];
  va_end(ap);
  free(cfmt);
}

@end /* NSMutableString(XSFormatting) */

// var args wrappers

int xs_sprintf(char *str, const char *format, ...) {
  va_list ap;
  int result;
  va_start(ap, format);
  result = xs_vsprintf(str, format, ap);
  va_end(ap);
  return result;
}
int xs_snprintf(char *str, size_t size, const char *format, ...) {
  va_list ap;
  int result;
  va_start(ap, format);
  result = xs_vsnprintf(str, size, format, ap);
  va_end(ap);
  return result;
}

/* static linking */

void __link_NSString_Formatting(void) {
  __link_NSString_Formatting();
}
