// $Id: NSURL+misc.m,v 1.2 2002/11/15 09:07:10 helge Exp $

#include "NSURL+misc.h"
#include "common.h"

@implementation NSURL(misc)

- (NSString *)stringByAddingFragmentAndQueryToPath:(NSString *)_path {
  NSString *lFrag, *lQuery;
  
  if ([self isFileURL])
    return _path;
  
  lFrag   = [self fragment];
  lQuery  = [self query];
  
  if ((lFrag != nil) || (lQuery != nil)) {
    NSMutableString *ms;
    
    ms = [NSMutableString stringWithCapacity:([_path length] + 32)];
    
    [ms appendString:_path];
    
    if (lFrag) {
      [ms appendString:@"#"];
      [ms appendString:lFrag];
    }
    if (lQuery) {
      [ms appendString:@"?"];
      [ms appendString:lQuery];
    }
    return ms;
  }
  else
    return _path;
}

- (NSString *)stringValueRelativeToURL:(NSURL *)_base {
  NSString *relPath;
  
  if (_base == self || _base == nil) {
    relPath = [[self path] urlPathRelativeToSelf];
    return [self stringByAddingFragmentAndQueryToPath:relPath];
  }

  /* check whether we are already marked relative to _base .. */
  if ([self baseURL] == _base)
    return [self stringByAddingFragmentAndQueryToPath:[self relativePath]];
  
  /* check whether we are in the same path namespace ... */
  if (![self isInSameNamespaceWithURL:_base])
    /* need to return full URL */
    return [self absoluteString];
  
  relPath = [[self path] urlPathRelativeToPath:[_base path]];
  return [self stringByAddingFragmentAndQueryToPath:relPath];
}

static BOOL isEqual(id o1, id o2) {
  if (o1 == o2) return YES;
  if (o1 == nil || o2 == nil) return NO;
  return [o1 isEqual:o2];
}

- (BOOL)isInSameNamespaceWithURL:(NSURL *)_url {
  if (_url == nil)  return NO;
  if (_url == self) return YES;
  if ([self isFileURL] && [_url isFileURL]) return YES;
  if ([self baseURL] == _url) return YES;
  if ([_url baseURL] == self) return YES;
  
  if (![[self scheme] isEqualToString:[_url scheme]])
    return NO;
  
  if (!isEqual([self host], [_url host]))
    return NO;
  if (!isEqual([self port], [_url port]))
    return NO;
  if (!isEqual([self user], [_url user]))
    return NO;
  
  return YES;
}

@end /* NSURL */

@implementation NSString(URLPathProcessing)

- (NSString *)urlPathRelativeToSelf {
  /*
    eg:                "/a/b/c.html"
    should resolve to: "c.html"
    
    Directories are a bit more difficult, eg:
      "/a/b/c/"
    is resolved to
      "../c/"
  */
  NSString *p;
  NSString *lp;
  
  p  = self;
  lp = [p lastPathComponent];
  
  p = ([p hasSuffix:@"/"])
    ? [NSString stringWithFormat:@"../%@/", p]
    : lp;
  return p;
}

- (NSString *)urlPathRelativeToRoot {
  NSString *p;
  
  p = self;
  
  if ([p isEqualToString:@"/"])
    /* don't know better ... what is root-relative-to-root ? */
    return @"/";
  
  if ([p length] == 0) {
    NSLog(@"%s: invalid path (length 0), using /: %@",
          __PRETTY_FUNCTION__, self);
    return @"/";
  }
  
  /* this is the same like the absoltute path, only without a leading "/" .. */
  return [p substringFromIndex:1];
}

- (NSString *)urlPathRelativeToPath:(NSString *)_base {
  /*
    This can be used for URLs in the same namespace. It should
    never return and absolute path (it only does in error conditions).
  */
  
  if (_base == nil || [_base length] == 0) {
    NSLog(@"%s: invalid base (nil or length 0), using absolute path '%@' ...",
          __PRETTY_FUNCTION__, self);
    return self;
  }
  
  if ([_base isEqualToString:@"/"])
    return [self urlPathRelativeToRoot];
  if ([_base isEqualToString:self])
    return [self urlPathRelativeToSelf];
  
  if ([self hasPrefix:_base]) {
    /*
      the whole base URI is prefix of our URI:
        case a)
          b: "/a/b/c"
          s: "/a/b/c/d"
          >: "c/d"
        case b)
          b: "/a/b/c/"
          s: "/a/b/c/d"
          >: "d"
        case c)
          b: "/a/b/c"
          s: "/a/b/ccc/d"
          >: "ccc/d"
        
      b=s is already catched above and s is guaranteed to be
      longer than b.
    */
    unsigned blen;
    NSString *result;
    
    blen = [_base length];
    
    if ([_base characterAtIndex:(blen - 1)] == '/') {
      /* last char of 'b' is '/' => case b) */
      result = [self substringFromIndex:blen];
    }
    else {
      /*
        last char of 'b' is not a slash (either case a) or case c)),
        both are handled in the same way (search last / ...)
      */
      NSRange  r;
        
      r = [_base rangeOfString:@"/" options:NSBackwardsSearch];
      if (r.length == 0) {
        NSLog(@"%s: invalid base, found no '/': '%@' !",
              __PRETTY_FUNCTION__, _base);
        result = self;
      }
      else {
        /* no we have case b) ... */
        result = [self substringFromIndex:(r.location + 1)];
      }
    }
    return result;
  }
  else {
    NSString *prefix;
    unsigned plen;
    
    prefix = [self commonPrefixWithString:_base options:0];
    plen = [prefix length];
    
    if (plen == 0) {
      NSLog(@"%s: invalid strings, no common prefix ...: '%@' and '%@' !",
              __PRETTY_FUNCTION__, self, _base);
      return self;
    }
    
    if (plen == 1) {
      /*
        common prefix is root. That is, nothing in common:
          b: "/a/b"
          s: "/l"
          >: "../l"
          
          b: "/a/b/"
          s: "/l"
          >: "../../l"
      */
      
      if ([prefix characterAtIndex:0] != '/') {
        NSLog(@"%s: invalid strings, common prefix '%@' is not '/': "
              @"'%@' and '%@' !",
              __PRETTY_FUNCTION__, self, _base, prefix);
      }
      
#warning to be completed ...
      return self;
    }
    
#warning to be completed ...
  }
  
  return self;
}

@end /* NSString(URLPathProcessing) */
