// $Id: NGImap4Context.m,v 1.19 2003/02/14 15:19:30 jan Exp $

#include "NGImap4Context.h"
#include "NGImap4Client.h"
#include "NGImap4Folder.h"
#include "NGImap4ServerRoot.h"
#include "NGImap4Support.h"
#include "NGImap4Functions.h"
#import "common.h"

@interface NGImap4Context(Private)
- (void)initializeSentFolder;
- (void)initializeTrashFolder;
- (void)initializeDraftsFolder;
- (void)initializeInboxFolder;
- (void)initializeServerRoot;
@end

@implementation NGImap4Context

+ (id)imap4ContextWithURL:(NSURL *)_url {
  return [[[self alloc] initWithURL:_url] autorelease];
}
+ (id)imap4ContextWithConnectionDictionary:(NSDictionary *)_connection {
  return [[[self alloc] initWithConnectionDictionary:_connection] autorelease];
}

- (id)initWithConnectionDictionary:(NSDictionary *)_connection {
  if ((self = [super init])) {
    self->connectionDictionary = [_connection copy];
    self->folderForRefresh = [[NSMutableArray alloc] initWithCapacity:512];
    self->syncMode = NO;
  }
  return self;
}
- (id)initWithURL:(NSURL *)_url {
  NSMutableDictionary *md;
  id                  tmp;

  if (_url == nil) {
    [self release];
    return nil;
  }
  md = [NSMutableDictionary dictionaryWithCapacity:4];
  if ((tmp = [_url host])) [md setObject:tmp forKey:@"host"];
  if ((tmp = [_url port])) [md setObject:tmp forKey:@"port"];
  if ((tmp = [_url user])) [md setObject:tmp forKey:@"login"];
  if ((tmp = [_url password])) [md setObject:tmp forKey:@"passwd"];
  
  if ([[_url scheme] isEqualToString:@"imaps"])
    [md setObject:[NSNumber numberWithBool:YES] forKey:@"SSL"];
  
  return [self initWithConnectionDictionary:md];
}

- (void)dealloc {
  [self->url release];
  [self->client removeFromResponseNotification:self];
  [self->connectionDictionary release];
  [self->client           release];
  [self->folderForRefresh release];
  [self->serverName       release];       
  [self->serverKind       release];       
  [self->serverVersion    release];    
  [self->serverSubVersion release]; 
  [self->serverTag        release];        
  [self->lastException    release];

  self->selectedFolder = nil; /* not retained */  
  self->trashFolder    = nil; /* not retained */  
  self->draftsFolder   = nil; /* not retained */  
  self->sentFolder     = nil; /* not retained */
  self->inboxFolder    = nil; /* not retained */
  self->serverRoot     = nil; /* not retained */

  [self->capability release];
  
  [super dealloc];
}

- (NSException *)lastException {
  return self->lastException;
}

- (void)setLastException:(NSException *)_exception {
  ASSIGN(self->lastException, _exception);
}

- (void)resetLastException {
  [self->lastException release]; self->lastException = nil;
}

- (NGImap4Client *)client {
  if (self->client == nil)
    [self openConnection];
  return self->client;
}

- (NSURL *)url {
  if (!self->url) {
    self->url = [[NSURL alloc]
                        initWithScheme:
                        [[self->connectionDictionary objectForKey:@"SSL"]
                                                     boolValue]?@"imaps":@"imap"
                        host:[self->connectionDictionary objectForKey:@"host"]
                        path:nil];
  }
  return self->url;
}

- (BOOL)isSelectedFolder:(NGImap4Folder *)_folder {
  return [self->selectedFolder isEqual:_folder];
}

- (void)setSelectedFolder:(NGImap4Folder *)_folder {
  self->selectedFolder = _folder;
}

- (BOOL)registerAsSelectedFolder:(NGImap4Folder *)_folder {
  if (self->selectedFolder != _folder) {
    if ([_folder noselect] == NO) {
      self->selectedFolder = _folder;
      if (![_folder selectImmediately:YES])
        return NO;
    }
    else {
      return NO;
    }
  }
  return YES;
}

- (BOOL)removeSelectedFolder:(NGImap4Folder *)_folder {
  if (self->selectedFolder == _folder)
    self->selectedFolder = nil;
  return YES;
}

- (BOOL)openConnection {
  NSString *login;
  NSString *passwd;
  NSString *host;

  login  = [self->connectionDictionary objectForKey:@"login"];
  passwd = [self->connectionDictionary objectForKey:@"passwd"];
  host   = [self->connectionDictionary objectForKey:@"host"];

  [self resetSpecialFolders];

  if ((login == nil) || (passwd == nil) || (host == nil)) {
    id exc;
    
    exc = [[NGImap4ConnectionException alloc] 
	    initWithFormat:
	      @"missing login, passwd or host in connection-dictionary <%@>",
	      self->connectionDictionary];
    ASSIGN(self->lastException, exc);
    [exc release];
    return NO;
  }
  if (self->client == nil) {
    self->client = [[NGImap4Client alloc] initWithHost:host];
    [self->client registerForResponseNotification:self];
    [self->client setContext:self];
  }
  if (![[self->client isConnected] boolValue]) {
    NSDictionary *res;
    
    [self resetLastException];
    
    res = [self->client openConnection];
    if (!_checkResult(self, res, __PRETTY_FUNCTION__))
      return NO;
    
    [self->serverName       release]; self->serverName       = nil;
    [self->serverKind       release]; self->serverKind       = nil;
    [self->serverVersion    release]; self->serverVersion    = nil;
    [self->serverSubVersion release]; self->serverSubVersion = nil;
    [self->serverTag        release]; self->serverTag        = nil;
    
    self->serverName = [[res objectForKey:@"server"] retain];
    self->serverKind = [[res objectForKey:@"serverKind"] retain];
    
    if (self->serverKind != nil) {
      self->serverVersion    = [[res objectForKey:@"version"] retain];
      self->serverSubVersion = [[res objectForKey:@"subversion"] retain];
      self->serverTag        = [[res objectForKey:@"tag"] retain];
    }
    
    [self resetLastException];
    res = [self->client login:login password:passwd];
    if (!_checkResult(self, res, __PRETTY_FUNCTION__))
      return NO;
  }
  {
    NSDictionary *res;

    [self->capability release]; self->capability = nil;
    res = [self->client capability];

    if (!_checkResult(self, res, __PRETTY_FUNCTION__))
      return NO;

    self->capability = [[res objectForKey:@"capability"] retain];
  }

  self->canSort  = -1;
  self->canQuota = -1;
  
  return YES;
}

- (BOOL)closeConnection {
  [self->client closeConnection];
  [self resetSpecialFolders];
  [self->capability release]; self->capability = nil;
  self->canSort = -1;
  self->canQuota = -1;

  
  return YES;
}

/*"
**  NGImap4ResponseReceiver protocol
**  If the NGImap4Context receive a response-notification it
**  updates the selected folder
"*/

- (void)responseNotificationFrom:(NGImap4Client *)_client
   response:(NSDictionary *)_dict
{
  if ([[_dict objectForKey:@"result"] boolValue] == NO) {
    id       exc;
    NSString *str;
    
    if (!(str = [_dict objectForKey:@"reason"]))
      str = @"Response failed";
    
    exc = [[NGImap4ResponseException alloc] initWithFormat:str];
    [(NSException *)exc setUserInfo:_dict];
    ASSIGN(self->lastException, exc);
    return;
  }
  if (self->selectedFolder != nil) {
    [self->selectedFolder processResponse:_dict];
  }
}

- (id)trashFolder {
  if (self->trashFolder == nil)
    [self initializeTrashFolder];
  return self->trashFolder;
}
- (void)setTrashFolder:(NGImap4Folder *)_folder {
  self->trashFolder = _folder;
}

- (id)sentFolder {
  if (self->sentFolder == nil)
    [self initializeSentFolder];
  return self->sentFolder;
}
- (void)setSentFolder:(NGImap4Folder *)_folder {
  self->sentFolder = _folder;
}

- (id)draftsFolder {
  if (self->draftsFolder == nil)
    [self initializeDraftsFolder];
  return self->draftsFolder;
}
- (void)setDraftsFolder:(NGImap4Folder *)_folder {
  self->draftsFolder = _folder;
}

- (id)inboxFolder {
  if (self->inboxFolder == nil)
    [self initializeInboxFolder];
  return self->inboxFolder;
}

- (id)serverRoot {
  if (self->serverRoot == nil)
    [self initializeServerRoot];
  return self->serverRoot;
}

- (void)initializeServerRoot {
  if (self->serverRoot == nil) {
    /*
      Note: serverRoot is not retained by NGImap4Context to avoid a
      retain cycle. This is why the object is autoreleased immediatly
      after creation (the usercode uses -serverRoot to access the result).
    */
    [self resetSpecialFolders];
    self->serverRoot = [[NGImap4ServerRoot alloc]
                                           initServerRootWithContext:self];
    self->serverRoot = [self->serverRoot autorelease];
  }
}

- (NGImap4Folder *)_getFolderWithName:(NSString *)_name {
  NSEnumerator  *enumerator;
  NGImap4Folder *folder;

  if (self->serverRoot == nil)
    [self initializeServerRoot];

  enumerator = [[self->serverRoot subFolders] objectEnumerator];
  while ((folder = [enumerator nextObject])) {
    NSString *name;

    name = [[folder name] lowercaseString];

    if ([name isEqualToString:[_name lowercaseString]]) {
      return folder;
    }
  }
  if ([[_name lowercaseString] isEqual:@"inbox"]) {
    [self resetLastException];
    [self->client subscribe:_name];
    if (self->lastException != nil) {
      [[self serverRoot] resetFolder]; /* because of client operation */
      [self->serverRoot createSubFolderWithName:_name];
      [self->lastException release]; self->lastException = nil;
    }
    [self resetSpecialFolders];
  }
  else {
    if ([[self inboxFolder] noinferiors]) {
      /* try to create Sent/Trash/Drafts in root */
      [self resetLastException];
      [self->client subscribe:_name];
      if (self->lastException != nil) {
        [[self serverRoot] resetFolder];  /* because of client operation */
        [self->serverRoot createSubFolderWithName:_name];
        [self->lastException release]; self->lastException = nil;
      }
      [[self serverRoot] resetFolder];  /* because of client operation */
      [self resetSpecialFolders];
    }
    else {
      /* take a look in inbox */
      NSString *absoluteName;

      folder = [[self inboxFolder] subFolderWithName:
                                   [_name lowercaseString]
                                   caseInsensitive:YES];
      if (folder != nil)
        return folder;
      
      absoluteName = [[[self inboxFolder] absoluteName]
                             stringByAppendingPathComponent:_name];
      [self resetLastException];
      [self->client subscribe:absoluteName];
      if (self->lastException != nil) {
        [[self inboxFolder] resetFolder];
        [[self inboxFolder] createSubFolderWithName:_name];
        [self->lastException release]; self->lastException = nil;
      }
    }
  }
  return nil;
}

- (NSString *)sentFolderName {
  static NSString *SentFolderName = nil;

  if (SentFolderName == nil) {
    SentFolderName = [[[NSUserDefaults standardUserDefaults]
                                      stringForKey:@"ImapSentFolderName"]
                                       retain];

    if (!SentFolderName)
      SentFolderName = @"Sent";
  }
  return SentFolderName;
}

- (NSString *)trashFolderName {
  static NSString *TrashFolderName = nil;

  if (TrashFolderName == nil) {
    TrashFolderName = [[[NSUserDefaults standardUserDefaults]
                                      stringForKey:@"ImapTrashFolderName"]
                                       retain];

    if (!TrashFolderName)
      TrashFolderName = @"Trash";
  }
  return TrashFolderName;
}

- (NSString *)draftsFolderName {
  static NSString *DraftsFolderName = nil;

  if (DraftsFolderName == nil) {
    DraftsFolderName = [[[NSUserDefaults standardUserDefaults]
                                      stringForKey:@"ImapDraftsFolderName"]
                                       retain];

    if (!DraftsFolderName)
      DraftsFolderName = @"Drafts";
  }
  return DraftsFolderName;
}

- (void)initializeSentFolder {
  if ((self->sentFolder = [self _getFolderWithName:
                                [self sentFolderName]]) == nil)
    self->sentFolder = [self _getFolderWithName:
                             [self sentFolderName]];
  if (self->sentFolder == nil)
    NSLog(@"WARNING[%s]: Couldn't find/create sentFolder", __PRETTY_FUNCTION__);
}

- (void)initializeTrashFolder {
  if ((self->trashFolder = [self _getFolderWithName:
                                 [self trashFolderName]]) == nil)
    self->trashFolder = [self _getFolderWithName:[self trashFolderName]];
  if (self->trashFolder == nil)
    NSLog(@"WARNING[%s]: Couldn't find/create trashFolder", __PRETTY_FUNCTION__);
}

- (void)initializeDraftsFolder {
  if ((self->draftsFolder = [self _getFolderWithName:
                                  [self draftsFolderName]]) == nil)
    self->draftsFolder = [self _getFolderWithName:
                               [self draftsFolderName]];
  if (self->draftsFolder == nil)
    NSLog(@"WARNING[%s]: Couldn't find/create draftsFolder", __PRETTY_FUNCTION__);
}

- (void)initializeInboxFolder {
  if ((self->inboxFolder = [self _getFolderWithName:@"Inbox"]) == nil)
    self->inboxFolder = [self _getFolderWithName:@"Inbox"];
  
  if (self->inboxFolder == nil)
    NSLog(@"WARNING[%s]: Couldn't find/create inbox", __PRETTY_FUNCTION__);
}

- (NGImap4Folder *)folderWithName:(NSString *)_name {
  return [self folderWithName:_name caseInsensitive:NO];
}

- (NGImap4Folder *)folderWithName:(NSString *)_name
  caseInsensitive:(BOOL)_caseIn {
  NSEnumerator      *enumerator;
  id                obj;
  id<NGImap4Folder> f;

  [self resetLastException];
  enumerator = [[_name componentsSeparatedByString:@"/"] objectEnumerator];
  f          = [self serverRoot];
  
  while ((obj = [enumerator nextObject])) {
    if ([obj length] > 0)
      f = [f subFolderWithName:obj caseInsensitive:_caseIn];
  }
  return self->lastException ? nil : f;
}

- (BOOL)createFolderWithPath:(NSString *)_name {
  NSEnumerator      *enumerator;
  id<NGImap4Folder> f1, f2; 
  NSString          *name;

  [self resetLastException];
  
  enumerator = [[_name componentsSeparatedByString:@"/"] objectEnumerator];
  f1         = [self serverRoot];
  f2         = nil;
  while ((name = [enumerator nextObject])) {
    if ((f2 = [f1 subFolderWithName:name caseInsensitive:YES]) == nil)
      break;
    f1 = f2;
  }
  if (name != nil) {
    do {
      if ([f1 createSubFolderWithName:name] == NO)
        break;
      f1 = [f1 subFolderWithName:name caseInsensitive:YES];
    } while ((name = [enumerator nextObject]));
  }
  return self->lastException ? NO : YES;
}

- (void)resetSpecialFolders {
  self->sentFolder   = nil;
  self->trashFolder  = nil;
  self->draftsFolder = nil;
  self->inboxFolder  = nil;
  self->serverRoot   = nil;
}

- (NSArray *)newMessages {
  NSEnumerator   *enumerator;
  NGImap4Folder  *f;
  NSMutableArray *result;
  EOQualifier    *qual;

  [self resetLastException];
  
  qual   = [EOQualifier qualifierWithQualifierFormat:@"flags = \"recent\""];
  result = [NSMutableArray array];
  
  [self->inboxFolder status];
  if ([self->inboxFolder hasNewMessagesSearchRecursiv:NO] == YES) {
    NSArray *array;

    array = [self->inboxFolder messagesForQualifier:qual];
    if (array != nil)
      [result addObjectsFromArray:array];
  }
  enumerator = [self->folderForRefresh objectEnumerator];  
  while ((f = [enumerator nextObject])) {
    [f status];
    if ([f hasNewMessagesSearchRecursiv:NO] == YES) {
    NSArray *array;

    array = [self->inboxFolder messagesForQualifier:qual];
    if (array != nil)
      [result addObjectsFromArray:array];
    }
  }
  return self->lastException ? nil : result;
}

- (BOOL)hasNewMessages {
  NSEnumerator  *enumerator;
  NGImap4Folder *f;
  BOOL          result;

  [self resetLastException];
  
  [self->inboxFolder status];
  if ([self->inboxFolder hasNewMessagesSearchRecursiv:NO] == YES)
    return YES;

  result     = NO;
  enumerator = [self->folderForRefresh objectEnumerator];
  
  while ((f = [enumerator nextObject])) {
    [f status];
    if ([f hasNewMessagesSearchRecursiv:NO] == YES) {
      result = YES;
      break;
    }
  }
  return self->lastException ? NO : result;
}

- (NSString *)host {
  return [self->connectionDictionary objectForKey:@"host"];
}
- (NSString *)login {
  return [self->connectionDictionary objectForKey:@"login"];
}

- (BOOL)registerForRefresh:(NGImap4Folder *)_folder {
  [self->folderForRefresh addObject:_folder];
  return YES;
}

- (BOOL)removeFromRefresh:(NGImap4Folder *)_folder {
  [self->folderForRefresh removeObject:_folder];
  return YES;
}

- (BOOL)removeAllFromRefresh {
  [self->folderForRefresh removeAllObjects];
  return YES;
}

- (BOOL)refreshFolder {
  NSEnumerator  *enumerator;
  NGImap4Folder *f;
  BOOL          refreshInbox = NO;

  if ([self lastException] != nil)
    return NO;
  
  enumerator = [self->folderForRefresh objectEnumerator];

  [self resetLastException];

  while ((f = [enumerator nextObject])) {
    if ([f isEqual:self->inboxFolder])
      refreshInbox = YES;
    
    [f status];
  }
  
  if (!refreshInbox)
    [self->inboxFolder status];
    
  return self->lastException ? NO : YES;
}

- (id)serverName {
  return self->serverName;
}
- (id)serverKind {
  return self->serverKind;
}
- (id)serverVersion {
  return self->serverVersion;
}
- (id)serverSubVersion {
  return self->serverSubVersion;
}
- (id)serverTag {
  return self->serverTag;
}

// Syncronize

- (void)resetSync {
  if (self->syncMode == YES) 
    [self->serverRoot resetSync];
  else {
    NSLog(@"WARNING: resetSync has no effect if syncMode == NO");
  }
}

- (BOOL)isInSyncMode {
  return self->syncMode;
}

- (void)enterSyncMode {
  self->syncMode = YES;
  [self resetSync];
}

- (void)leaveSyncMode {
  self->syncMode = NO;
}

/* description */

- (NSString *)description {
  NSMutableString *ms;
  NSString *tmp;

  ms = [NSMutableString stringWithCapacity:64];

  [ms appendFormat:@"<0x%08X[%@]:", self, NSStringFromClass([self class])];

  if ((tmp = [self host]))
    [ms appendFormat:@" host=%@", tmp];
  if ((tmp = [self login]))
    [ms appendFormat:@" login=%@", tmp];
  
  [ms appendFormat:@" server='%@'/%@/%@.%@/%@",
        [self serverName],
        [self serverKind],
        [self serverVersion],
        [self serverSubVersion],
        [self serverTag]];
  
  if (self->syncMode)
    [ms appendString:@" syncmode"];
  
  [ms appendString:@">"];

  return ms;
}

@end /* NGImap4Context */

@implementation NGImap4Context(Capability)

- (BOOL)canSort {
  if (self->capability == nil) {
    if (![self openConnection])
      return NO;
  }
  if (self->canSort == -1) {
    self->canSort =
      ([self->capability containsObject:@"sort"])? 1 : 0;
  }
  return self->canSort;
}

- (BOOL)canQuota {
  if (self->capability == nil) {
    if (![self openConnection])
      return NO;
  }
  if (self->canQuota == -1) {
    self->canQuota =
      ([self->capability containsObject:@"quota"])? 1 : 0;
  }
  return self->canQuota;
}
@end /* NGImap4Context(Capability) */


