//$Id: NGImap4DataSource.m,v 1.6 2003/04/01 13:50:03 gnustep Exp $

#include "NGImap4DataSource.h"
#include <NGStreams/NGSocketExceptions.h>
#include "common.h"

@interface NGImap4DataSource(PrivateMethodes)
- (NSArray *)fetchMessages;
@end

@interface EOQualifier(PrivateMethodes)
- (NSString *)qualifierDescription;
- (BOOL)isImap4UnseenQualifier;
@end

@implementation EOQualifier(IMAPAdditions)

- (BOOL)isImap4UnseenQualifier {
  return [[self qualifierDescription] rangeOfString:@"unseen"].length > 0
    ? YES : NO;
}

@end /* EOQualifier(IMAPAdditions) */

@implementation NGImap4DataSource

- (id)init {
  if ((self = [super init])) {
    NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
    
    [nc addObserver:self
        selector:@selector(mailsWereDeleted:)
        name:@"LSWImapMailWasDeleted"
        object:nil];

    [nc addObserver:self
        selector:@selector(folderWasMoved:)
        name:@"LSWImapMailFolderWasDeleted"
        object:nil];
    
    [nc addObserver:self
        selector:@selector(flagsWereChanged:)
        name:@"LSWImapMailFlagsChanged"
        object:nil];
  }
  return self;
}
- (id)initWithFolder:(NGImap4Folder *)_folder {
  if ((self = [self init])) {
    [self setFolder:_folder];
  }
  return self;
}

- (void)dealloc {
  [[NSNotificationCenter defaultCenter] removeObserver:self];
  [self->folder   release];
  [self->fspec    release];
  [self->messages release];
  [self->oldUnseenMessages release];
  [super dealloc];
}

/* operations */

- (void)folderWasMoved:(id)_obj {
  [self->folder   release]; self->folder   = nil;
  [self->messages release]; self->messages = nil;
  [self->oldUnseenMessages release]; self->oldUnseenMessages = nil;
}

- (void)mailsWereDeleted:(id)_obj {
  [self->messages          release]; self->messages          = nil;
  [self->oldUnseenMessages release]; self->oldUnseenMessages = nil;
}

- (void)flagsWereChanged:(id)_obj {
  [self postDataSourceChangedNotification];
}

- (NSArray *)fetchObjects {
  if (self->messages == nil) {
    NSAutoreleasePool *pool;
    NSArray *tmp = nil;
    
    pool = [[NSAutoreleasePool alloc] init];
    
    tmp = (self->folder) ? [self fetchMessages] : [NSArray array];
    
    ASSIGN(self->messages, tmp);
    [pool release];
  }

  return self->messages;
}

- (void)clear {
  ASSIGN(self->messages, (id)nil);
}

- (void)setFolder:(NGImap4Folder *)_folder {
  ASSIGN(self->folder,            _folder);
  ASSIGN(self->messages,          (id)nil);
  ASSIGN(self->oldUnseenMessages, (id)nil);
  
  [self postDataSourceChangedNotification];
}
- (NGImap4Folder *)folder {
  return self->folder;
}

- (void)setFetchSpecification:(EOFetchSpecification *)_fetchSpec {
  if (![_fetchSpec isEqual:self->fspec]) {
    ASSIGN(self->fspec, _fetchSpec);
    ASSIGN(self->messages, (id)nil);
    
    [self postDataSourceChangedNotification];
  }
}
- (EOFetchSpecification *)fetchSpecification {
  return self->fspec;
}

- (int)oldExists {
  return self->oldExists;
}
- (int)oldUnseen {
  return self->oldUnseen;
}

// --- private methodes -----------------------------------------------------

- (NSException *)handleImapException:(NSException *)_exception {
  NSString  *rawResp = @"RawResponse";
  NSString  *respRes = @"ResponseResult";
  NSString  *descr   = @"description";
  
  if ([_exception isKindOfClass:[NGImap4ResponseException class]]) {
    NSString *str = [[[[_exception userInfo]
                                   objectForKey:rawResp]
                                   objectForKey:respRes]
                                   objectForKey:descr];
    fprintf(stderr, "%s", [str cString]);
  }    
  else if ([_exception isKindOfClass:[NGIOException class]]) {
    fprintf(stderr, "%s", [[_exception reason] cString]);
  }
  else if ([_exception isKindOfClass:[NGImap4Exception class]]) {
    fprintf(stderr, "%s", [[_exception description] cString]);
  }
  else {
    fprintf(stderr, "\n");
    return _exception;
  }
  fprintf(stderr, "\n");
  return nil;
}

- (NSArray *)fetchMessages {
  EOQualifier *qualifier;
  NSArray     *result  = nil;
  
  qualifier = [[self fetchSpecification] qualifier];
  
  NS_DURING {
    if (![qualifier isImap4UnseenQualifier]) {
      [self->oldUnseenMessages release];
      self->oldUnseenMessages = nil;
    }
    
    self->oldExists = [self->folder exists];
    self->oldUnseen = [self->folder unseen];
    
    result = (qualifier == nil)
      ? [self->folder messages]
      : [self->folder messagesForQualifier:qualifier];
  }
  NS_HANDLER {
    [[self handleImapException:localException] raise];
    result = [NSArray array];
  }
  NS_ENDHANDLER;

  /* process unseen hh?? */
  
  if ([qualifier isImap4UnseenQualifier]) {
    if (self->oldUnseenMessages) { // this implies, the folder hasn't changed
      NSMutableSet *set = nil;
      
      set = [[NSMutableSet alloc] initWithArray:result];
      [set addObjectsFromArray:self->oldUnseenMessages];
      result = [[[set allObjects] retain] autorelease];
      [set release];
    }
    [self->oldUnseenMessages release]; self->oldUnseenMessages = nil;
    ASSIGN(self->oldUnseenMessages, result);
  }
  
  /* sort results */

  {
    NSArray *sortOrderings;

    if ((sortOrderings = [[self fetchSpecification] sortOrderings]))
      result = [result sortedArrayUsingKeyOrderArray:sortOrderings];
  }
  
  /* return result */
  
  return result;
}

@end /* NGImap4DataSource */
