// $Id: NGSieveClient.m,v 1.4 2003/01/07 11:38:27 helge Exp $

#include <unistd.h>

#include "NGSieveClient.h"
#include "NGImap4Support.h"
#include "NGImap4ResponseParser.h"
#include "NSString+Imap4.h"
#include "common.h"
#include <sys/time.h>

@interface NGSieveClient(Private)

- (NGHashMap *)processCommand:(NSString *)_command;
- (NGHashMap *)processCommand:(NSString *)_command logText:(NSString *)_txt;

- (void)sendCommand:(NSString *)_command;
- (void)sendCommand:(NSString *)_command logText:(NSString *)_txt;

- (NSMutableDictionary *)normalizeResponse:(NGHashMap *)_map;
- (NSMutableDictionary *)normalizeOpenConnectionResponse:(NGHashMap *)_map;
- (NSDictionary *)login;

@end

/*"
**  An implementation of an Imap4 client
**
** A folder name always looks like an absolute filename (/inbox/doof) 
**
"*/

@implementation NGSieveClient

static BOOL ProfileImapEnabled = NO;

static NSNumber *YesNumber     = nil;
static NSNumber *NoNumber      = nil;
static int       LOG_PASSWORD  = -1;

+ (void)initialize {
  if (LOG_PASSWORD == -1)
    LOG_PASSWORD = [[NSUserDefaults standardUserDefaults]
                                    boolForKey:@"SieveLogPassword"]?1:0;
  if (YesNumber == nil)
    YesNumber = [[NSNumber numberWithBool:YES] retain];
  if (NoNumber == nil)
    NoNumber = [[NSNumber numberWithBool:NO] retain];
  
  ProfileImapEnabled = [[NSUserDefaults standardUserDefaults]
                                         boolForKey:@"ProfileImapEnabled"];
}

+ (id)clientWithAddress:(id<NGSocketAddress>)_address {
  return
    [[(NGSieveClient *)[self alloc] initWithAddress:_address] autorelease];
}

+ (id)clientWithHost:(id)_host {
  return [[[self alloc] initWithHost:_host] autorelease];
}

- (id)initWithHost:(id)_host {
  NGInternetSocketAddress *a;
  a = [NGInternetSocketAddress addressWithPort:143 onHost:_host];
  return [self initWithAddress:a];
}

/**"
 ** designated initializer
"**/

- (id)initWithAddress:(id<NGSocketAddress>)_address {
  if ((self = [super init])) {
    self->address = [_address retain];
    
    self->debug =
      [[NSUserDefaults standardUserDefaults] boolForKey:@"ImapDebugEnabled"];
    
  }
  return self;
}

- (void)dealloc {
  [self->text     release];
  [self->address  release];
  [self->socket   release];
  [self->parser   release];
  [self->login    release];
  [self->password release];
  [super dealloc];
}

/* equality */

- (BOOL)isEqual:(id)_obj {
  if (_obj == self)
    return YES;
  if ([_obj isKindOfClass:[NGSieveClient class]])
    return [self isEqualToSieveClient:_obj];
  return NO;
}

- (BOOL)isEqualToSieveClient:(NGSieveClient *)_obj {
  if (_obj == self) return YES;
  return [(NGInternetSocketAddress *)[_obj address]
                                     isEqualToAddress:self->address];
}

/* accessors */

- (id<NGActiveSocket>)socket {
  return self->socket;
}

- (id<NGSocketAddress>)address {
  return self->address;
}

/* connection */

/*"
** Opens a connection to given Host.
"*/

- (NSDictionary *)openConnection {
  struct timeval tv;
  double         ti = 0.0;
  
  if (ProfileImapEnabled) {
    gettimeofday(&tv, NULL);
    ti =  (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0);
  }

  [self->socket release]; self->socket = nil;
  [self->parser release]; self->parser = nil;
  [self->text   release]; self->text   = nil;
  
  self->socket =
    [[NGActiveSocket socketConnectedToAddress:self->address] retain];
  self->text   = [[NGCTextStream alloc] initWithSource:self->socket];
  self->parser = [[NGImap4ResponseParser alloc] initWithStream:self->socket];

  /* receive greeting from server without tag-id */

  if (ProfileImapEnabled) {
    gettimeofday(&tv, NULL);
    ti = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0) - ti;
    fprintf(stderr, "[%s] <openConnection> : time needed: %4.4fs\n",
           __PRETTY_FUNCTION__, ti < 0.0 ? -1.0 : ti);    
  }
  return [self normalizeOpenConnectionResponse:
               [self->parser parseSieveResponse]];
}

/*"
** Check whether stream is already open (could be closed because server-timeout)
"*/

- (NSNumber *)isConnected {
  return [NSNumber numberWithBool:
                     (self->socket == nil)
                     ? NO : [(NGActiveSocket *)self->socket isAlive]];
}

/*"
** Close a consisting connection.
"*/

- (void)closeConnection {
  [self->socket close];
  [self->socket release]; self->socket = nil;
  [self->text   release]; self->text   = nil; 
  [self->parser release]; self->parser = nil;
}

/*"
** login with plaintext password authenticating
"*/

- (NSDictionary *)login:(NSString *)_login password:(NSString *)_passwd {
  NSZone *z = [self zone];

  if ((_login == nil) || (_passwd == nil))
    return nil;

  [self->login    release];
  [self->password release];
  
  self->login    = [_login  copyWithZone:z];
  self->password = [_passwd copyWithZone:z];
  return [self login];
}

- (void)reconnect {
  [self closeConnection];  
  [self openConnection];
  [self login];
}

- (NSDictionary *)login {
  NGHashMap *map  = nil;
  NSData    *auth;
  char      *buf;
  int       bufLen, logLen;


  logLen = [self->login cStringLength];
  bufLen = (logLen*2) + [self->password cStringLength] +2;

  buf = calloc(sizeof(char), bufLen);

  sprintf(buf, "%s %s %s", 
          [self->login cString], [self->login cString],
          [self->password cString]);
  
  buf[logLen] = '\0';
  buf[logLen*2+1] = '\0';
  
  auth = [NSData dataWithBytesNoCopy:buf length:bufLen];
  auth = [auth dataByEncodingBase64];

  if (LOG_PASSWORD == 1) {
    map = [self processCommand:[NSString stringWithFormat:
                                         @"AUTHENTICATE \"PLAIN\" {%d+}\r\n%s",
                                         [auth length], [auth bytes]]];
  }
  else {
    map = [self processCommand:[NSString stringWithFormat:
                                         @"AUTHENTICATE \"PLAIN\" {%d+}\r\n%s",
                                         [auth length], [auth bytes]]
                logText:@"AUTHENTICATE \"PLAIN\" {%d+}\r\nLOGIN:PASSWORD\r\n"];
  }
  return [self normalizeResponse:map];
}

/*
** logout from the connected host and close the connection
*/

- (NSDictionary *)logout {
  NGHashMap *map = [self processCommand:@"logout"];

  [self closeConnection];
  return [self normalizeResponse:map];
}

- (NSDictionary *)getScript:(NSString *)_scriptName {
  [self notImplemented:_cmd];
  return nil;
}
- (NSDictionary *)putScript:(NSString *)_name script:(NSString *)_script {
  NGHashMap *map;

  if (![_name length]) {
    NSLog(@"%s: missing script-name", __PRETTY_FUNCTION__);
    return nil;
  }

  if (![_script length]) {
    NSLog(@"%s: missing script", __PRETTY_FUNCTION__);
    return nil;
  }
  
  map = [self processCommand:
              [NSString stringWithFormat:
                        @"PUTSCRIPT \"%@\" {%d+}\r\n%@\r\n",
                        _name, [_script length], _script]];
  return [self normalizeResponse:map];
}

- (NSDictionary *)setActiveScript:(NSString *)_name {
  NGHashMap *map;

  if (![_name length]) {
    NSLog(@"%s: missing script-name", __PRETTY_FUNCTION__);
    return nil;
  }
  map = [self processCommand:
              [NSString stringWithFormat:@"SETACTIVE \"%@\"\r\n", _name]];
  return [self normalizeResponse:map];
}

- (NSDictionary *)deleteScript:(NSString *)_name {
  NGHashMap *map;

  if (![_name length]) {
    NSLog(@"%s: missing script-name", __PRETTY_FUNCTION__);
    return nil;
  }
  map = [self processCommand:
              [NSString stringWithFormat:@"DELETESCRIPT \"%@\"\r\n", _name]];
  return [self normalizeResponse:map];
}
- (NSDictionary *)listScript:(NSString *)_script {
  [self notImplemented:_cmd];
  return nil;
}



/*
** Filter for all responses
**       result  : NSNumber (response result)
**       exists  : NSNumber (number of exists mails in selectet folder
**       recent  : NSNumber (number of recent mails in selectet folder
**       expunge : NSArray  (message sequence number of expunged mails
                             in selectet folder)
*/

- (NSMutableDictionary *)normalizeResponse:(NGHashMap *)_map {
  NSMutableDictionary *result  = nil;

  NSAssert(_map, @"missing map parameter !");

  result  = [NSMutableDictionary dictionaryWithCapacity:32];

  [result setObject:_map forKey:@"RawResponse"];

  if ([[_map objectForKey:@"ok"] boolValue]) {
    [result setObject:YesNumber forKey:@"result"];
  }
  else {
    [result setObject:NoNumber forKey:@"result"];    
  }
  return result;
}

/*
** filter for open connection
*/

- (NSDictionary *)normalizeOpenConnectionResponse:(NGHashMap *)_map {
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if (([[[_map objectEnumeratorForKey:@"ok"] nextObject] boolValue])) {
    NSString *tmp;

    if ((tmp = [_map objectForKey:@"implementation"]))
      [result setObject:tmp forKey:@"server"];

    if ((tmp = [_map objectForKey:@"sieve"]))
      [result setObject:tmp forKey:@"capabilities"];
  }
  return result;
}


 
/*
** filter for list
**       list : NSDictionary (folder name as key and flags as value)
*/

- (NSString *)description {
  return [NSString stringWithFormat:@"<Imap4Client[0x%08X]: socket=%@>",
                     (unsigned)self, [self socket]];
}

/* Private Methods */

- (NGHashMap *)processCommand:(NSString *)_command logText:(NSString *)_txt {
  NGHashMap *map          = nil;
  BOOL      repeatCommand = NO;
  int       repeatCnt     = 0;

  struct timeval tv;
  double         ti = 0.0;

  if (ProfileImapEnabled) {
    gettimeofday(&tv, NULL);
    ti =  (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0);
    fprintf(stderr, "{");
  }
  do {
    if (repeatCommand == YES) {
      if (repeatCnt > 1) {
        fprintf(stderr, "Reconnect to %s, wait for %d seconds\n",
                [[self->address description] cString], repeatCnt * 2);
        sleep(repeatCnt * 4);
      }
      repeatCnt++;
      [self reconnect];
      repeatCommand = NO;
    }
    
    NS_DURING {
      [self sendCommand:_command logText:_txt];
      map = [self->parser parseSieveResponse];
    }
    NS_HANDLER {
      if (repeatCnt > 3) {
        [localException raise];
      }
      else if ([localException isKindOfClass:[NGIOException class]] == YES) {
        fprintf(stderr, "WARNING: Got Exception try to restore connection "
                "localException: %s\n", [[localException description] cString]);
        repeatCommand = YES;
      }
      else if ([localException isKindOfClass:
                               [NGImap4ParserException class]] == YES) {
        fprintf(stderr, "WARNING: Got Parser-Exception try to restore connection "
                "localException: %s\n", [[localException description] cString]);
        repeatCommand = YES;
      }
      else {
        [localException raise];
      }
    }
    NS_ENDHANDLER;    
  }
  while (repeatCommand == YES);
  
  if (ProfileImapEnabled) {
    gettimeofday(&tv, NULL);
    ti = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0) - ti;
    fprintf(stderr, "}[%s] <Send Command> : time needed: %4.4fs\n",
           __PRETTY_FUNCTION__, ti < 0.0 ? -1.0 : ti);    
  }
  
  return map;
}

- (NGHashMap *)processCommand:(NSString *)_command {
  return [self processCommand:_command logText:_command];
}

- (void)sendCommand:(NSString *)_command logText:(NSString *)_txt
{
  NSString *command       = nil;

  command = _command;

  if (debug == YES) {
    fprintf(stderr, "C: %s\n", [_txt cString]);
  }
  [self->text writeString:command];
  [self->text writeString:@"\r\n"];
  [self->text flush];
}

- (void)sendCommand:(NSString *)_command {
  [self sendCommand:_command logText:_command];
}

@end /* NGSieveClient */
