// $Id: NGMimeMessageParser.m,v 1.19 2003/02/17 17:32:09 jan Exp $

#include "NGMimeMessageParser.h"
#include "NGMimeMessage.h"
#include "common.h"

@interface NGMimeMessageParserDelegate : NSObject
@end

@implementation NGMimeMessageParserDelegate

static Class NGMimeMessageParserClass = NULL; 

- (id)parser:(id)_parser parseHeaderField:(NSString *)_field data:(NSData *)_data
{
  id v = nil;

  if (NGMimeMessageParserClass == NULL) {
    NGMimeMessageParserClass = [NGMimeMessageParser class];
  }
  if ([_parser isKindOfClass:NGMimeMessageParserClass] == NO) {
    NGMimeMessageParser *parser = nil;

    parser = [[NGMimeMessageParserClass alloc] init];
    v = [parser valueOfHeaderField:_field data:_data];
    [parser release]; parser = nil;
  }
  return v;
}

- (id<NGMimeBodyParser>)parser:(NGMimePartParser *)_parser
  bodyParserForPart:(id<NGMimePart>)_part
{
  id                   ctype;
  NGMimeType           *contentType;

  ctype = [_part contentType];
  
  contentType = ([ctype isKindOfClass:[NGMimeType class]])
    ? ctype
    : [NGMimeType mimeType:[ctype stringValue]];
  
  if ([[contentType type] isEqualToString:@"message"] &&
      [[contentType subType] isEqualToString:@"rfc822"]) {
    return [[[NGMimeRfc822BodyParser alloc] init] autorelease];
  }
  return nil;
}


@end /* NGMimeMessageParserDelegate */

@implementation NGMimeMessageParser

static Class NSStringClass = Nil;

+ (int)version {
  return 3;
}
+ (void)initialize {
  NSAssert2([super version] == 3,
            @"invalid superclass (%@) version %i !",
            NSStringFromClass([self superclass]), [super version]);
  NSStringClass = [NSString class];
}

- (id)init {
  if ((self = [super init])) {
    [self setDelegate:[NGMimeMessageParserDelegate new]];
  }
  return self;
}

/* factory */

- (id<NGMimePart>)producePartWithHeader:(NGHashMap *)_header {
  return [NGMimeMessage messageWithHeader:_header];
}

/* header field specifics */

- (id)valueOfHeaderField:(NSString *)_name data:(id)_data {
  // check data for 8-bit headerfields (RFC 2047 (MIME PART III))
  
  /* check whether we got passed a string ... */
  if ([_data isKindOfClass:NSStringClass]) {
    NSLog(@"%s: WARNING unexpected class for headerfield %@ (value %@)",
          __PRETTY_FUNCTION__, _name, _data);
    return [super valueOfHeaderField:_name data:_data];
  }
  _data = [_data decodeQuotedPrintableValueOfMIMEHeaderField:_name];
  return [super valueOfHeaderField:_name data:_data];
}

@end /* NGMimeMessageParser */

@implementation NSData(MimeQPHeaderFieldDecoding)

- (id)decodeQuotedPrintableValueOfMIMEHeaderField:(NSString *)_name {
  // check data for 8-bit headerfields (RFC 2047 (MIME PART III))
  static Class NGMimeTypeClass = Nil;
  enum {
    NGMimeMessageParser_quoted_start   = 1,
    NGMimeMessageParser_quoted_charSet = 2,
    NGMimeMessageParser_quoted_qpData  = 3,
    NGMimeMessageParser_quoted_end     = 4
  } status = NGMimeMessageParser_quoted_start;
  unsigned int        length;
  const unsigned char *bytes, *firstEq;
  BOOL foundQP = NO;

  if (NSStringClass   == Nil) NSStringClass   = [NSString class];
  if (NGMimeTypeClass == Nil) NGMimeTypeClass = [NGMimeType class];
  
  length = [self length];
  
  /* check whether the string is long enough to be quoted etc */
  if (length <= 6)
    return self;
  
  /* check whether the string contains QP tokens ... */
  bytes = [self bytes];
  if ((firstEq = memchr(bytes, '=', length)) == NULL)
    return self;
  
  /* process data ... (quoting etc) */
  {
    unsigned char *buffer;
    unsigned int  bufLen, maxBufLen;
    NSString      *charset;
    BOOL          appendLC;
    int           cnt, tmp;
    unsigned char encoding;
    
    buffer = calloc(length + 13, sizeof(unsigned char));
    
    maxBufLen             = length + 3;
    buffer[maxBufLen - 1] = '\0';
    bufLen                = 0;
    
    encoding = 0;
    tmp      = -1;
    appendLC = YES;      
    charset  = nil;
    status   = NGMimeMessageParser_quoted_start;

    /* copy data up to first '=' sign */
    if ((cnt = (firstEq - bytes)) > 0) {
      memcpy(buffer, bytes, cnt);
      bufLen += cnt;
    }
    
    for (; cnt < (length-1); cnt++) {
      appendLC = YES;      
      
      if (status == NGMimeMessageParser_quoted_start) {
        if ((bytes[cnt] == '=') && (bytes[cnt + 1] == '?')) { // found begin
          cnt++;
          status = NGMimeMessageParser_quoted_charSet;
        }
        else { // other char
          if (bytes[cnt + 1] != '=') {
            buffer[bufLen++] = bytes[cnt];
            buffer[bufLen++] = bytes[cnt+1];
            cnt++;
            if (cnt >= length - 1)
              appendLC = NO;
          }
          else {
            buffer[bufLen++] = bytes[cnt];
          }
        }
      }
      else if (status == NGMimeMessageParser_quoted_charSet) {
        if (tmp == -1)
          tmp = cnt;
	
        if (bytes[cnt] == '?') {
          charset = 
	    [NSStringClass stringWithCString:(bytes + tmp) length:cnt - tmp];
          tmp = -1;
	  
          if ((length - cnt) > 2) { 
	    // set encoding (eg 'q' for quoted printable)
            cnt++; // skip '?'
            encoding = bytes[cnt];
            cnt++; // skip encoding
            status = NGMimeMessageParser_quoted_qpData;
          }
          else { // unexpected end
            NSLog(@"WARNING: unexpected end of header");
            appendLC = NO;
            break;
          }
        }
      }
      else if (status == NGMimeMessageParser_quoted_qpData) {
        if (tmp == -1)
          tmp = cnt;
	
        if ((bytes[cnt] == '?') && (bytes[cnt + 1] == '=')) {
          NSData           *tmpData;
          NSString         *tmpStr;
	  unsigned int     tmpLen;
          NSStringEncoding enc;
	  
          tmpData = _rfc2047Decoding(encoding, bytes + tmp, cnt - tmp);
	  foundQP = YES;
          
	  enc = [NGMimeTypeClass stringEncodingForCharset:charset];
	  
	  /* 
	     create a temporary string for charset conversion ... 
	     Note: the headerfield is currently held in ISO Latin 1
	  */
          tmpStr = [[NSStringClass alloc] initWithData:tmpData encoding:enc];
	  tmpLen = [tmpStr cStringLength];
	  
	  if ((tmpLen + bufLen) < maxBufLen) {
	    [tmpStr getCString:(buffer + bufLen)];
	    bufLen += tmpLen;
	  }
	  else {
	    NSLog(@"ERROR[%s]: quoted data to large --> ignored %@",
		  __PRETTY_FUNCTION__, tmpStr);
	  }
          [tmpStr release]; tmpStr = nil;
	  
          tmp = -1;
          cnt++;
          appendLC = NO;
          status = NGMimeMessageParser_quoted_start;
        }
      }
    }
    
    if (appendLC == YES) {
      buffer[bufLen] = bytes[cnt];
      bufLen++;
    }
    buffer[bufLen] = '\0';
    
    if (buffer) {
      id data;
      
      if (foundQP) {
	data = [NSStringClass stringWithCString:buffer length:bufLen];
	if (data == nil) {
	  NSLog(@"%s: got no string for buffer '%s', length '%i' !", 
		__PRETTY_FUNCTION__,
		buffer, bufLen);
	}
      }
      else {
#if DEBUG && 0
	if (bufLen != length) {
	  NSLog(@"%s: parsed length %i and orginal length %i do not match, "
		@"even though no quoted printable encoding was found !",
		__PRETTY_FUNCTION__,
		bufLen, length);
	}
#endif
	data = self;
      }
      
      free(buffer);

      return data;
    }
    else
      return self;
  }
}

@end /* NSData(MimeQPHeaderFieldDecoding) */

@implementation NGMimeRfc822BodyParser

+ (int)version {
  return 2;
}
+ (void)initialize {
  NSAssert2([super version] == 2,
            @"invalid superclass (%@) version %i !",
            NSStringFromClass([self superclass]), [super version]);
}

- (id)parseBodyOfPart:(id<NGMimePart>)_part data:(NSData *)_data
  delegate:(id)_d
{
  id<NGMimePart> body;
  id             parser; // NGMimeMessageParser

  parser = [[NGMimeMessageParser alloc] init];
  body = [parser parsePartFromData:_data];
  [parser release]; parser = nil;
  
  return body;
}

@end /* NGMimeRfc822BodyParser */
