// $Id: NGMimeFileData.m,v 1.5 2003/01/29 15:26:26 helge Exp $

#include "NGMimeFileData.h"
#include "common.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

@implementation NGMimeFileData

- (id)initWithPath:(NSString *)_path removeFile:(BOOL)_remove
{
  if ((self = [super init])) {
    if (![[NSFileManager defaultManager] fileExistsAtPath:_path]) {
      NSLog(@"ERROR[%s]: missing file at path %@", __PRETTY_FUNCTION__, _path);
      [self release];
      return nil;
    }
    ASSIGN(self->path, _path);
    self->removeFile = _remove;
    self->length     = -1;
  }
  return self;
}

- (id)initWithBytes:(const void*)_bytes
    length:(unsigned int)_length
{
  NSString *filename = nil;
  int      fd;

  static NSString      *TmpPath = nil;
  static NSProcessInfo *Pi      = nil;

  if (TmpPath == nil) {
    TmpPath = [[NSUserDefaults standardUserDefaults]
                               stringForKey:@"NGMimeBuildMimeTempDirectory"];
    if (!TmpPath) {
      TmpPath = @"/tmp/";
    }
    TmpPath = [[TmpPath stringByAppendingPathComponent:@"Skyrix"] retain];
    
  }
  if (Pi == nil)
    Pi = [[NSProcessInfo processInfo] retain];

  filename = [Pi temporaryFileName:TmpPath];

  fd = open([filename fileSystemRepresentation],
            O_WRONLY | O_CREAT | O_TRUNC, 0600);
  if (fd == -1) {
    fprintf(stderr, "Could not open file for writing %s: %s\n",
            [filename fileSystemRepresentation], strerror(errno));
    RELEASE(self);
    return nil;
  }
  if (write(fd, _bytes, _length) != _length) {
    fprintf(stderr, "Failed to write %i bytes to %s: %s\n",
            _length, [filename fileSystemRepresentation], strerror(errno));
    close(fd);
    RELEASE(self);
    return nil;
  }
  return [self initWithPath:filename removeFile:YES];
}

- (void)dealloc {
  if (self->removeFile) {
    [[NSFileManager defaultManager]
                    removeFileAtPath:self->path handler:nil];
  }
  [self->path release];
  [super dealloc];
}

- (NSData *)_data {
  return [NSData dataWithContentsOfMappedFile:self->path];
}

- (id)copyWithZone:(NSZone*)zone {
    return RETAIN(self);
}

- (const void*)bytes {
  return [[self _data] bytes];
}

- (unsigned int)length {
  if (self->length == -1) {
    self->length = [[[[NSFileManager defaultManager]
                                     fileAttributesAtPath:self->path
                                     traverseLink:NO]
                                     objectForKey:NSFileSize] intValue];
  }
  return self->length;
}

- (NSString *)description {
  return [NSString stringWithFormat:@"<0x%08X[%@]: path=%@>",
                   self, NSStringFromClass([self class]), self->path];
}

- (BOOL)appendDataToFileDesc:(int)_fd {
  NGFileStream *fs;
  int  bufCnt = 8192;
  char buffer[bufCnt];
  BOOL result;
  int  fileLen;

  if (![[NSFileManager defaultManager] isReadableFileAtPath:self->path]) {
    NSLog(@"ERROR[%s] missing file at path %@", __PRETTY_FUNCTION__, 
          self->path);
    return NO;
  }
  
  fileLen = [self length];
  result  = YES;
  fs      = [[NGFileStream alloc] initWithPath:self->path];

  if (![fs openInMode:@"r"]) {
    NSLog(@"%s: couldn`t open file stream ... %@",
          __PRETTY_FUNCTION__, self->path);
    RELEASE(fs); fs = nil;
    return NO;
  }

  NS_DURING {
    int read;
    int alreadyRead;

    alreadyRead = 0;
    
    read = (bufCnt > (fileLen - alreadyRead)) ? fileLen - alreadyRead : bufCnt;
        
    while ((read = [fs readBytes:buffer count:read])) {
      alreadyRead += read;
      if (write(_fd, buffer, read) != read) {
        fprintf(stderr, "%s: Failed to write %i bytes to file\n",
                __PRETTY_FUNCTION__, read);
        result = NO;
        break;
      }
      if (alreadyRead == fileLen)
        break;
    }
  }
  NS_HANDLER {
    printf("got exceptions %s\n", [[localException description] cString]);
    if (![localException isKindOfClass:[NGEndOfStreamException class]]) {
      RELEASE(fs); fs = nil;
      result = NO;
    }
  }
  NS_ENDHANDLER;
  RELEASE(fs); fs = nil;
  return result;
}
@end /* NGMimeFileData */
