// $Id: NGMimeHeaderFields.m,v 1.3 2003/01/20 10:45:56 helge Exp $

#include "NGMimeHeaderFields.h"
#include "NGMimeUtilities.h"
#include "common.h"

NGMime_DECLARE NSString *NGMimeContentDispositionInlineType     = @"inline";
NGMime_DECLARE NSString *NGMimeContentDispositionAttachmentType = @"attachment";
NGMime_DECLARE NSString *NGMimeContentDispositionFormType       = @"form-data";

@implementation NGMimeContentDispositionHeaderField

+ (int)version {
  return 2;
}

- (id)initWithString:(NSString *)_value {
  unsigned len = [_value cStringLength];
  char *buf;
  char *cstr;
  
  if (len == 0) {
    [self release];
    NSLog(@"WARNING(%s): no value for disposition header value !",
          __PRETTY_FUNCTION__);
    return nil;
  }

  buf = cstr = malloc(len + 1);
  
  [_value getCString:buf]; buf[len] = '\0';
  
  if ((self = [super init])) {
    const char *start = NULL;
    
    // skip leading spaces
    while (isRfc822_LWSP(*cstr)) cstr++;
    if (*cstr == '\0') {
      self = [self autorelease];
      NSLog(@"WARNING(%s): no value for disposition header value !",
            __PRETTY_FUNCTION__);
      free(buf);
      return nil;
    }

    start = cstr;
    while ((*cstr != ';') && (*cstr != '\0') && !isRfc822_LWSP(*cstr))
      cstr++;

    if (cstr <= start) {
      self = [self autorelease];
      NSLog(@"WARNING(%s): found no type in disposition header value (%@) !",
            __PRETTY_FUNCTION__, _value);
      free(buf);
      return nil;
    }
    self->type = [NSString stringWithCString:start length:(cstr - start)];
    start = NULL;
    self->type = [self->type lowercaseString];
    self->type = [self->type retain];
    if (self->type == nil) {
      self = [self autorelease];
      NSLog(@"WARNING(%s): found no type in disposition header value (%@) !",
            __PRETTY_FUNCTION__, _value);
      free(buf);
      return nil;
    }

    if (*cstr != '\0') { // read parameters
      do {
        NSString *name  = nil;
        NSString *value = nil;

        // consume end of previous entry (spaces and ';')
        while ((*cstr == ';') || isRfc822_LWSP(*cstr))
          cstr++;

        // read name
        start = cstr;
        while ((*cstr != '=') && (*cstr != '\0') && !isRfc822_LWSP(*cstr))
          cstr++;

        if ((start < cstr) && (*cstr != '\0')) {
          name = [NSString stringWithCString:start length:(cstr - start)];
          name = [name lowercaseString];
          start = NULL;

          // skip after '=' and LSWP
          while ((*cstr != '=') && (*cstr != '\0')) cstr++;
          
          if (*cstr == '=') { // value follows
            cstr++; // skip '='
            
            while ((*cstr != '\0') && isRfc822_LWSP(*cstr)) cstr++;
            if (*cstr == '"') { // quoted value
              cstr++; // skip quote
              if (*cstr == '\0') continue;

              start = cstr;
              while ((*cstr != '\0') && (*cstr != '"')) cstr++;

              value = [NSString stringWithCString:start length:(cstr - start)];
              start = NULL;

              if (*cstr != '\0') cstr++; // skip closing quote
            }
            else if (*cstr != '\0') { // unquoted value
              start = cstr;

              while ((*cstr != '\0') && !isRfc822_LWSP(*cstr) && (*cstr != ';'))
                cstr++;

              value = [NSString stringWithCString:start length:(cstr - start)];
              start = NULL;
            }
          }
        }

        if ((name != nil) && (value != nil)) {
          if (self->parameters == nil) {
            self->parameters = [[NSMutableDictionary allocWithZone:[self zone]]
                                                     initWithCapacity:4];
          }
          [(NSMutableDictionary *)self->parameters setObject:value forKey:name];
        }
      }
      while (*cstr != '\0');
    }
  }
  free(buf);
  return self;
}

- (id)init {
  return [self initWithString:nil];
}

- (void)dealloc {
  [self->type       release];
  [self->parameters release];
  [super dealloc];
}

/* accessors */

- (NSString *)type {
  return self->type;
}

/* parameters */

- (NSString *)name {
  return [self->parameters objectForKey:@"name"];
}
- (NSString *)filename {
  NSString *fn;

  fn = [self->parameters objectForKey:@"filename"];

  if (![fn isNotNull])
    fn = nil;
  
  if (![fn length]) {
    fn = [self name];
  }
  return fn;
}

- (NSString *)valueOfParameterWithName:(NSString *)_name {
  return [self->parameters objectForKey:_name];
}

- (BOOL)valueNeedsQuotes:(NSString *)_parameterValue {
  char *cstr =(char *)[_parameterValue cString];
  while (*cstr) {
    if (isMime_SpecialByte(*cstr))
      return YES;

    if (*cstr == 32)
      return YES;
    
    cstr++;
  }
  return NO;
}

- (NSString *)parametersAsString {
  NSEnumerator *names = [self->parameters keyEnumerator];

  if (names) {
    NSMutableString *result = [NSMutableString stringWithCapacity:64];
    NSString *name;

    while ((name = [names nextObject])) {
      id value = [[parameters objectForKey:name] stringValue];
      [result appendString:@"; "];
      [result appendString:name];
      [result appendString:@"="];
      if ([self valueNeedsQuotes:value]) {
        [result appendString:@"\""];
        [result appendString:value];
        [result appendString:@"\""];
      }
      else
        [result appendString:value];
    }
    return result;
  }
  else
    return nil;
}

- (NSString *)stringValue {
  NSMutableString *str = [NSMutableString stringWithCapacity:20];
  [str appendString:type];
  {
    NSString *paras = [self parametersAsString];
    if (paras) [str appendString:paras];
  }
  return str;
}

/* description */

- (NSString *)description {
  NSMutableString *d = [[NSMutableString alloc] init];

  [d appendFormat:@"<%@[0x%08X]: type=%@",
       NSStringFromClass([self class]), self, self->type];

  if (self->parameters)
    [d appendFormat:@" parameters=%@", self->parameters];

  [d appendString:@">"];
  return AUTORELEASE(d);
}

@end /* NGMimeContentDispositionHeaderField */
