/* 
   NGByteBuffer.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Jan Reichmann (jan.reichmann@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGByteBuffer.m,v 1.7 2003/01/07 11:36:47 helge Exp $

#import "NGByteBuffer.h"
#include "common.h"
#include <sys/time.h>

typedef struct NGByteBufferLA {
  unsigned char byte;
  char          isEOF:1;
  char          isFetched:1;
} LA_NGByteBuffer;

static BOOL ProfileByteBuffer = NO;

@implementation NGByteBuffer

+ (void)initialize {
  static BOOL didInit = NO;

  if (!didInit) {
    didInit = YES;
    ProfileByteBuffer = [[NSUserDefaults standardUserDefaults]
                                         boolForKey:
                                           @"ProfileByteBufferEnabled"];
  }
}

+ (int)version {
  return [super version] + 1;
}

+ (id)byteBufferWithSource:(id<NGStream>)_source la:(unsigned)_la {
  return [[[self alloc] initWithSource:_source la:_la] autorelease];
}

- (id)initWithSource:(id<NGStream>)_source la:(unsigned)_la {
  if ((self = [super initWithSource:_source])) {
    int size = 0;

    if (_la < 1) {
      [[[RangeException alloc]
                        initWithReason:@"lookahead depth is less than one"
                        size:1 index:_la]
                        raise];
    }

    // Find first power of 2 >= to requested size
    for (size = 2; size < _la; size *=2);
    
#if NeXT_Foundation_LIBRARY
    self->la = NSZoneMalloc([self zone], sizeof(LA_NGByteBuffer) * size);
#else
    self->la = NSZoneMallocAtomic([self zone],
                                  sizeof(LA_NGByteBuffer) * size);
#endif
    memset(self->la, 0, sizeof(LA_NGByteBuffer) * size);

    self->bufLen      = size;
    self->sizeLessOne = self->bufLen - 1;
    self->headIdx     = 0;
    self->wasEOF      = NO;
    if ([self->source respondsToSelector:@selector(methodForSelector:)]) {
      self->readByte = (int(*)(id, SEL))
        [(NSObject *)self->source methodForSelector:@selector(readByte)];
    }
    if ([self respondsToSelector:@selector(methodForSelector:)]) {
      self->laFunction = (int(*)(id, SEL, unsigned))
        [(NSObject *)self methodForSelector:@selector(la:)];
    }
  }
  return self;
}

- (id)init {
  [self release];
  [self doesNotRecognizeSelector:_cmd];
  return nil;
}

- (id)initWithSource:(id<NGStream>)_source {
  [self release];
  [self doesNotRecognizeSelector:_cmd];
  return nil;
}

- (id)initWithInputSource:(id<NGInputStream>)_source {
  [self release];
  [self doesNotRecognizeSelector:_cmd];
  return nil;
}

- (id)initWithOutputSource:(id<NGOutputStream>)_source {
  [self release];
  [self doesNotRecognizeSelector:_cmd];
  return nil;
}

- (void)dealloc {
  NSZoneFree([self zone], self->la);
  self->readByte   = NULL;
  self->laFunction = NULL;
  [super dealloc];
}

- (int)readByte {
  int byte = (self->laFunction == NULL)
    ? [self la:0]
    : self->laFunction(self, @selector(la:), 0);
  [self consume];
  return byte;
}

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  if (_len == 0)
    return 0;

  if (!(self->la[(self->headIdx & self->sizeLessOne)].isFetched)) {
    int byte = [self readByte];

    if (byte == -1)
      [NGEndOfStreamException raiseWithStream:self->source];

    ((char *)_buf)[0] = byte;
    return 1;
  }
  else {
    int  cnt    = 0;
    int  idxCnt = self->headIdx & sizeLessOne;
    char buffer[self->bufLen];

    while (self->la[idxCnt].isFetched && cnt < _len && cnt < bufLen) {
      buffer[cnt] = self->la[idxCnt].byte;
      cnt++;
      idxCnt = (cnt + self->headIdx) & sizeLessOne;
    }
    memcpy(_buf, buffer, cnt);
    [self consume:cnt];
    return cnt;
  }
  return 0;
}

- (int)la:(unsigned)_la {
  volatile unsigned result, idx;

  result = -1;
  *(&idx) = (_la + self->headIdx) & self->sizeLessOne;
  
  if (_la > self->sizeLessOne) {
    [[[RangeException alloc]
                      initWithReason:@"tried to look ahead too far"
                      size:self->bufLen
                      index:_la]
                      raise];
  }
  
  if (self->wasEOF) {
    result = (!self->la[idx].isFetched || self->la[idx].isEOF)
      ? -1 : self->la[idx].byte;
  }
  else {
    if (!(self->la[idx].isFetched)) {
      int i = 0;

      *(&i) = 0;
      for (i = 0;
           i < _la &&
           self->la[(self->headIdx + i) & self->sizeLessOne].isFetched;
           i++);

      /* If we should read more than 5 bytes, we take the time costs of an */
      /* exception handler */
      if ((_la - i + 1) <= 5) {
        while (i <= _la) {
#if DEBUG
          struct timeval tv;
          double         ti = 0.0;
#endif
          
          int byte = 0;

#if DEBUG
          if (ProfileByteBuffer) {
            gettimeofday(&tv, NULL);
            ti =  (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0);
          }
#endif
          byte = (self->readByte == NULL)
            ? [self->source readByte]
            : (int)self->readByte(self->source, @selector(readByte));

#if DEBUG
          if (ProfileByteBuffer) {
            gettimeofday(&tv, NULL);
            ti = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0) - ti;
            if (ti > 0.01) {
              fprintf(stderr, "[%s] <read bytes from stream> : time "
                      "needed: %4.4fs\n",
                     __PRETTY_FUNCTION__, ti < 0.0 ? -1.0 : ti);
            }
          }
#endif
    
          if (byte == -1) {  // EOF was reached
            self->wasEOF = YES;
            break;
          }
          else {
            int ix = (self->headIdx + i) & self->sizeLessOne;
            self->la[ix].byte      = byte;
            self->la[ix].isFetched = 1;
          }
          i++;
        }
      }
      else {
        BOOL readStream = YES;
        NSException *exc = nil;
        
        while (readStream == YES) {
          int  cntReadBytes  = 0;
          int  cnt           = 0;
          int  desiredBytes  = _la - i+1;
          char *tmpBuffer;
          
          tmpBuffer = malloc(desiredBytes + 2);

          cntReadBytes = (self->readBytes == NULL)
            ? [self->source readBytes:tmpBuffer count:desiredBytes]
            : self->readBytes(self->source, @selector(readBytes:count:),
                              tmpBuffer, desiredBytes);
          
          if (cntReadBytes == NGStreamError) {
            exc = [[self->source lastException] retain];
            break;
          }
          else {
            if (cntReadBytes == desiredBytes)
              readStream = NO;

            cnt = 0;
            while (cntReadBytes > 0) {
              int ix = (self->headIdx + i) & self->sizeLessOne;
              self->la[ix].byte      = tmpBuffer[cnt];
              self->la[ix].isFetched = 1;
              i++;
              cnt++;
              cntReadBytes--;
            }
          }
          
          if (tmpBuffer) free(tmpBuffer);
        }
        if (exc) {
          if ([exc isKindOfClass:[NGEndOfStreamException class]] == NO) {
            [self setLastException:exc];
            return NGStreamError;
          }
          self->wasEOF = YES;
        }
      }
      if (self->wasEOF)
        while (i <= _la) {
          self->la[(self->headIdx + i) & self->sizeLessOne].isEOF = YES;
          i++;
        }
    }
    result = (self->la[idx].isEOF) ? -1 : self->la[idx].byte;
  }
  return result;
}

- (void)consume {
  int idx = self->headIdx & sizeLessOne;
  
  if (!(self->la[idx].isFetched)) {
    (self->laFunction == NULL)
      ? [self la:0]
      : self->laFunction(self, @selector(la:), 0);
  }
  self->la[idx].isFetched = 0;
  self->headIdx++;
}

- (void)consume:(unsigned)_cnt {
  while (_cnt > 0) {
    int idx = self->headIdx & sizeLessOne;
    
    if (!(self->la[idx].isFetched))
      (self->laFunction == NULL)
        ? [self la:0]
        : self->laFunction(self, @selector(la:), 0);

    self->la[idx].isFetched = 0;
    self->headIdx++;
    _cnt--;
  }
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:@"<%@[0x%08X] source=%@ mode=%@ la=%d",
                     NSStringFromClass([self class]),
                     (unsigned)self,
                     self->source,
                     [self modeDescription],
                     self->bufLen
                   ];
}

@end /* NGByteBuffer */
