/* 
   NGCharBuffer.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Jan Reichmann (jan.reichmann@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGCharBuffer.m,v 1.3 2003/01/07 11:36:47 helge Exp $

#include "NGCharBuffer.h"
#include "common.h"

typedef struct NGCharBufferLA {
  unichar character;
  char    isEOF:1;
  char    isFetched:1;
} LA_NGCharBuffer;

@implementation NGCharBuffer

+ (id)charBufferWithSource:(id<NGTextStream>)_source la:(int)_la {
  return [[[self alloc] initWithSource:_source la:_la] autorelease];
}

- (id)initWithSource:(id<NGTextStream>)_source la:(int)_la {
  if ((self = [super initWithSource:_source])) {
    int size = 0;

    if (_la < 1)
      [[[RangeException alloc]
                        initWithReason:@"LA is less than one"
                        size:1 index:_la]
                        raise];

    // Find first power of 2 >= to requested size
    for (size = 2; size < _la; size *=2);
    
#if NeXT_Foundation_LIBRARY
    self->la = NSZoneMalloc([self zone], sizeof(LA_NGCharBuffer) * size);
#else
    self->la = NSZoneMallocAtomic([self zone],
                                  sizeof(LA_NGCharBuffer) * size);
#endif
    memset(self->la, 0, sizeof(LA_NGCharBuffer) * size);

    self->bufLen      = size;
    self->sizeLessOne = self->bufLen - 1;
    self->headIdx     = 0;
    self->wasEOF      = NO;

    if ([self->source respondsToSelector:@selector(methodForSelector:)]) {
      self->readCharacter =
        [(NSObject *)self->source methodForSelector:@selector(readCharacter)];
    }
  }
  return self;
}

- (id)initWithSource:(id<NGTextStream>)_source {
  [self release];
  [self doesNotRecognizeSelector:_cmd];
  return nil;
}

- (id)init {
  [self release];
  [self doesNotRecognizeSelector:_cmd];
  return nil;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  NSZoneFree([self zone], self->la);
  self->readCharacter = NULL;
  [super dealloc];
}
#endif

- (unichar)readCharacter {
  int character = [self la:0];
  if (character < 1)
      [[[NGEndOfStreamException alloc] init] raise];
  [self consume];
  return character;
}

// LA

- (int)la:(int)_la {
  int result = -1;
  int idx    = (_la + self->headIdx) & self->sizeLessOne;

  idx = *(&idx);
  
  if (_la > self->sizeLessOne) {
    [[[RangeException alloc]
                      initWithReason:@"LA out of bounds"
                      size:self->bufLen
                      index:_la]
                      raise];
  }
  
  if (self->wasEOF) {
    result = (!self->la[idx].isFetched || self->la[idx].isEOF)
      ? -1
      : self->la[idx].character;
  }
  else {
    if (!self->la[idx].isFetched) {
      int i;

      *(&i) = 0;
      while ((i < _la) &&
             (self->la[(self->headIdx + i) & self->sizeLessOne].isFetched))
        i++;

      NS_DURING {
        while (i <= _la) {
          int     ix        = 0;
          unichar character = 0;

          if (self->readCharacter == NULL) 
            character = [self->source readCharacter];
          else
            character = (int)self->readCharacter(self->source,
                                                 @selector(readCharacter));
          ix = (self->headIdx + i) & self->sizeLessOne;      
          self->la[ix].character = character;
          self->la[ix].isFetched = 1;
          i++;
        }
      }
      NS_HANDLER {
        if ([localException isKindOfClass:[NGEndOfStreamException class]]) {
          while (i <= _la) {
            self->la[(self->headIdx + i) & self->sizeLessOne].isEOF = YES;
            i++;
            self->wasEOF = YES;        
          }
        }
        else {
          [localException raise];
        }
      }
      NS_ENDHANDLER;
    }
    result = (self->la[idx].isEOF) ? -1 : self->la[idx].character;
  }
  return result;
}

- (void)consume {
  int idx = self->headIdx & sizeLessOne;
  
  if (!(self->la[idx].isFetched))
    [self la:0];

  self->la[idx].isFetched = 0;
  self->headIdx++;
}

- (void)consume:(int)_cnt {
  while (_cnt > 0) {
    int idx = self->headIdx & sizeLessOne;
    
    if (!(self->la[idx].isFetched))
      [self la:0];

    self->la[idx].isFetched = 0;
    self->headIdx++;
    _cnt--;
  }
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:@"<%@[0x%08X] source=%@ la=%d",
                     NSStringFromClass([self class]),
                     (unsigned)self,
                     self->source,
                     self->bufLen
                   ];
}

@end
