/* 
   NGDataStream.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGDataStream.m,v 1.6 2003/01/07 11:36:47 helge Exp $

#include "common.h"
#include "NGDataStream.h"
#include "NGStreamExceptions.h"

@implementation NGDataStream

+ (int)version {
  return [super version] + 2;
}

+ (id)dataStream {
  return [self streamWithData:[NSMutableData dataWithCapacity:1024]];
}
+ (id)dataStreamWithCapacity:(int)_capacity {
  return [self streamWithData:[NSMutableData dataWithCapacity:_capacity]];
}

+ (id)streamWithData:(NSData *)_data {
  return [[[self alloc] initWithData:_data] autorelease];
}
- (id)initWithData:(NSData *)_data {
  if ((self = [super init])) {
    self->data     = [_data retain];
    self->position = 0;

    if ([self->data respondsToSelector:@selector(methodForSelector:)] == YES) {
      self->dataLength = (unsigned int(*)(id, SEL))
                         [self->data methodForSelector:@selector(length)];
      self->dataBytes  = (const void*(*)(id, SEL))
                         [self->data methodForSelector:@selector(bytes)];
    }
    else {
      self->dataLength = NULL;
      self->dataBytes  = NULL;
    }
    
    streamMode = [data isKindOfClass:[NSMutableData class]]
      ? NGStreamMode_readWrite
      : NGStreamMode_readOnly;
  }
  return self;
}

- (void)dealloc {
  [self->data release];
  self->dataLength = NULL;
  self->dataBytes  = NULL;
  [self->lastException release];
  [super dealloc];
}

// accessors

/* NGTextInputStream */

- (NSException *)lastException {
  return nil;
}
- (void)setLastException:(NSException *)_exception {
  ASSIGN(self->lastException, _exception);
}
- (void)resetLastException {
  ASSIGN(self->lastException, (id)nil);
}

- (NSData *)data {
  return self->data;
}

- (unsigned)availableBytes {
  // returns number of available bytes
  unsigned currentLength = 0;

  currentLength = (self->dataLength == NULL)
                ? [self->data length]
                : self->dataLength(self->data, @selector(length));

  return (currentLength == position)
    ? 0
    : (currentLength - position);
}

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  // throws
  //   NGStreamNotOpenException   when the stream is not open
  //   NGEndOfStreamException     when the end of the stream is reached
  unsigned currentLength = 0;

  currentLength = (self->dataLength == NULL)
                ? [self->data length]
                : self->dataLength(self->data, @selector(length));

  if (self->data == nil) {
    [self setLastException:
          [NGStreamNotOpenException exceptionWithStream:self reason:
                                    @"tried to read from a data stream "
                                    @"which was closed"]];
    return NGStreamError;
  }

  if (currentLength == position) {
    [self setLastException:
          [NGEndOfStreamException exceptionWithStream:self]];
    return NGStreamError;
  }
  {
    NSRange range;
    range.location = position;

    if ((position + _len) > currentLength)
      range.length = currentLength - position;
    else
      range.length = _len;

    [self->data getBytes:_buf range:range];

    position += range.length;
    return range.length;
  }
}

- (int)readByte {
  int currentLength = 0;
  int result        = 0;

  currentLength = (self->dataLength == NULL)
                ? [self->data length]
                : self->dataLength(self->data, @selector(length));
  
  if (currentLength == position)
    result = -1;
  else {
    register const unsigned char *p;
    p = (self->dataBytes == NULL)
      ? [self->data bytes]
      : self->dataBytes(self->data, @selector(bytes));
    result = p[position];
    position++;
  }
  return result;
}

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  // throws
  //   NGReadOnlyStreamException when the stream is not writeable
  //   NGStreamNotOpenException  when the stream is not open
  
  if (self->data == nil) {
    [self setLastException:
          [NGStreamNotOpenException exceptionWithStream:self reason:
                                    @"tried to write to a data stream "
                                    @"which was closed"]];
    return NGStreamError;
  }
  if (!NGCanWriteInStreamMode(streamMode)) {
    [self setLastException:
          [NGReadOnlyStreamException exceptionWithStream:self]];
    return NGStreamError;
  }
  [(NSMutableData *)self->data appendBytes:_buf length:_len];

  return _len;
}

- (BOOL)close {
  ASSIGN(self->lastException, (id)nil);
  RELEASE(self->data); self->data = nil;
  position   = 0;
  streamMode = NGStreamMode_undefined;
  return YES;
}

- (NGStreamMode)mode {
  return streamMode;
}
- (BOOL)isRootStream {
  return YES;
}

// NGPositionableStream

- (BOOL)moveToLocation:(unsigned)_location {
  position = _location;
  return YES;
}
- (BOOL)moveByOffset:(int)_delta {
  position += _delta;
  return YES;
}

// blocking ..

- (BOOL)wouldBlockInMode:(NGStreamMode)_mode {
  return NO;
}

- (id)retain {
  return [super retain];
}

@end /* NGDataStream */
