/* 
   NGLocalSocketAddress.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGLocalSocketAddress.m,v 1.3 2003/01/07 11:36:47 helge Exp $

#if !defined(WIN32) || defined(__CYGWIN32__)

#include "NGSocketExceptions.h"
#include "NGLocalSocketAddress.h"
#include "NGLocalSocketDomain.h"
#import <Foundation/NSString.h>

#include "config.h"

#if defined(__APPLE__)
#  include <sys/types.h>
#else
#  include <sys/un.h>
#endif

#if defined(HAVE_UNISTD_H) || defined(__APPLE__)
#  include <unistd.h>
#endif

#ifndef SUN_LEN
#define SUN_LEN(su) \
        (sizeof(*(su)) - sizeof((su)->sun_path) + strlen((su)->sun_path))
#endif

#include "common.h"

#ifndef AF_LOCAL
#  define AF_LOCAL AF_UNIX
#endif

#if defined(__WIN32__) && !defined(__CYGWIN32__)
static NSString *socketDirectoryPath = @"\\\\.\\pipe\\";
#else
static NSString *socketDirectoryPath = @"/tmp";
#endif

@implementation NGLocalSocketAddress

+ (id)addressWithPath:(NSString *)_path {
  return [[[self alloc] initWithPath:_path] autorelease];
}
+ (id)address {
  return [[[self alloc] init] autorelease];
}

- (id)initWithPath:(NSString *)_path {
  if ((self = [super init])) {
    self->address = calloc(1, sizeof(struct sockaddr_un));
    
    memset(self->address, 0, sizeof(struct sockaddr_un));
    
#if defined(__WIN32__) && !defined(__CYGWIN32__)
    self->path = [_path copyWithZone:[self zone]];
#else
    if ([_path cStringLength] >=
        sizeof(((struct sockaddr_un *)self->address)->sun_path)) {
      
      NSLog(@"LocalDomain name too long: maxlen=%i, len=%i, path=%@",
            sizeof(((struct sockaddr_un *)self->address)->sun_path),
            [_path cStringLength],
            _path);
      
      [[[InvalidArgumentException alloc] initWithReason:
           @"path to long as local domain socket address !"] raise];
      [self release];
      return nil;
    }
    
    ((struct sockaddr_un *)self->address)->sun_family =
      [[self domain] socketDomain];

    [_path getCString:((struct sockaddr_un *)self->address)->sun_path
           maxLength:sizeof(((struct sockaddr_un *)self->address)->sun_path)];
#endif
  }
  return self;
}

- (id)init {
  int      addressCounter = 0;
  NSString *newPath;
  
  newPath = [NSString stringWithFormat:@"_ngsocket_%08X_%08X_%03d",
                        getpid(), [NSThread currentThread], addressCounter];
  newPath = [socketDirectoryPath stringByAppendingPathComponent:newPath];

  return [self initWithPath:newPath];
}

- (id)initWithDomain:(id)_domain
  internalRepresentation:(void *)_representation
  size:(int)_length
{
  // this method is used by the address factory
  struct sockaddr_un *nun = _representation;
  NSString *path;

  path = (_length < 3)
    ? @""
    : [NSString stringWithCString:nun->sun_path];
  
  return [self initWithPath:path];
}

- (void)dealloc {
  if (self->address) free(self->address);
  [super dealloc];
}

/* accessors */

- (NSString *)path {
  const char *sp;

  sp = ((struct sockaddr_un *)self->address)->sun_path;
  if (strlen(sp) == 0)
    return @"";
  
  return [NSString stringWithCString:sp];
}

/* operations */

- (void)deletePath {
  const char *sp;
  
  sp = ((struct sockaddr_un *)self->address)->sun_path;
  if (strlen(sp) == 0)
    return;
  
  unlink(sp);
}

// NGSocketAddress protocol

- (void *)internalAddressRepresentation {
  return self->address;
}
- (int)addressRepresentationSize { // varies in length
  return SUN_LEN(((struct sockaddr_un *)self->address));
}
- (id)domain {
  return [NGLocalSocketDomain domain];
}

/* test for accessibility */

- (BOOL)canSendOnAddress {
  return (access(((struct sockaddr_un *)self->address)->sun_path, W_OK) == 0)
    ? YES : NO;
}
- (BOOL)canReceiveOnAddress {
  return (access(((struct sockaddr_un *)self->address)->sun_path, R_OK) == 0)
    ? YES : NO;
}

/* testing for equality */

- (BOOL)isEqualToAddress:(NGLocalSocketAddress *)_addr {
  return [[_addr path] isEqualToString:[self path]];
}

- (BOOL)isEqual:(id)_object {
  if (_object == self) return YES;
  if ([_object class] != [self class]) return NO;
  return [self isEqualToAddress:_object];
}

/* NSCopying */

- (id)copyWithZone:(NSZone *)_zone {
  /* socket addresses are immutable, just retain on copy ... */
  return [self retain];
}

/* NSCoding */

- (void)encodeWithCoder:(NSCoder *)_encoder {
  [_encoder encodeObject:[[NSHost currentHost] name]];
  [_encoder encodeObject:[self path]];
}

- (id)initWithCoder:(NSCoder *)_decoder {
  NSString *hostName = [_decoder decodeObject];
  NSString *path     = [_decoder decodeObject];

  NSAssert([path isKindOfClass:[NSString class]], @"path must be a string ..");

  if (![hostName isEqualToString:[[NSHost currentHost] name]]) {
    NSLog(@"unarchived local socket address on a different host, "
          @"encoded on %@, decoded on %@ (path=%@)",
          hostName, [[NSHost currentHost] name], path);
  }

  return [self initWithPath:path];
}

/* description */

- (NSString *)stringValue {
  NSString *p = [self path];
  return [p length] == 0 ? @"*" : (id)p;
}

- (NSString *)description {
  NSString *p = [self path];
  
  if ([p length] == 0)
    p = @"[no path]";

  return [NSString stringWithFormat:@"<0x%08X[%@]: %@>",
                     self, NSStringFromClass([self class]),
                     p];
}

@end

#endif /* !WIN32 */
