/* 
   NGPassiveSocket.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGPassiveSocket.m,v 1.3 2003/01/14 13:02:04 helge Exp $

#include "NGPassiveSocket.h"
#include "NGSocketExceptions.h"
#include "NGActiveSocket.h"
#include "NGSocket+private.h"

#if defined(__APPLE__)
#  include <sys/types.h>
#  include <sys/socket.h>
#endif

#if HAVE_SYS_ERRNO_H || defined(__APPLE__)
#  include <sys/errno.h>
#endif

#include "common.h"

@interface NGActiveSocket(privateMethods)

- (id)_initWithDescriptor:(int)_fd
  localAddress:(id<NGSocketAddress>)_local
  remoteAddress:(id<NGSocketAddress>)_remote;

@end

@implementation NGPassiveSocket

+ (id)socketBoundToAddress:(id<NGSocketAddress>)_address {
  volatile id sock = [[self alloc] initWithDomain:[_address domain]];

  if (sock != nil) {
    sock = AUTORELEASE(sock);
    [sock bindToAddress:_address];
  }
  return sock;
}

- (id)initWithDomain:(id<NGSocketDomain>)_domain { // designated initializer
  if ((self = [super initWithDomain:_domain])) {
    backlogSize = -1; // -1 means 'not listening'
    
    if ([NSThread isMultiThreaded])
      acceptLock = [[NSLock allocWithZone:[self zone]] init];
    else {
      acceptLock = nil;
      [[NSNotificationCenter defaultCenter]
                             addObserver:self
                             selector:@selector(taskNowMultiThreaded:)
                             name:NSWillBecomeMultiThreadedNotification
                             object:nil];
    }

    if (self->fd != NGInvalidSocketDescriptor) {
      int i_yes = 1;
      
      if (setsockopt(self->fd, SOL_SOCKET, SO_REUSEADDR,
                     (void *)&i_yes, sizeof(int)) != 0) {
        NSLog(@"WARNING: could not set SO_REUSEADDR option for socket %@: %s",
              self, strerror(errno));
      }
    }
  }
  return self;
}

- (void)dealloc {
  [[NSNotificationCenter defaultCenter]
                         removeObserver:self
                         name:NSWillBecomeMultiThreadedNotification
                         object:nil];
  
  [self->acceptLock release];
  [super dealloc];
}

- (void)taskNowMultiThreaded:(NSNotification *)_notification {
  if (acceptLock == nil) acceptLock = [[NSLock alloc] init];
}

// accessors

- (BOOL)isListening {
  return (backlogSize != -1);
}
- (BOOL)isOpen {
  return [self isListening];
}

- (id<NGSocketAddress>)localAddress {
  return localAddress;
}

- (int)socketType {
  return SOCK_STREAM;
}

/* operations */

#if defined(WIN32) && !defined(__CYGWIN32__)
- (NSString *)reasonForLastError {
  int errorCode = WSAGetLastError();

  switch (errorCode) {
    case WSAEBADF:
      return @"not a valid socket descriptor";
    case WSAENOTSOCK:
      return @"descriptor is not a socket descriptor";
    case WSAEOPNOTSUPP:
      return @"socket does not support listen";
    case WSAEINTR:
      return @"interrupted by signal";
    case WSAEMFILE:
      return @"descriptor table is full";

    default:
      return [NSString stringWithCString:strerror(errorCode)];
  }
}
#else
- (NSString *)reasonForLastError {
  int errorCode = errno;
  
  switch (errorCode) {
    case EBADF:
      return @"not a valid socket descriptor";
    case ENOTSOCK:
      return @"descriptor is not a socket descriptor";
    case EOPNOTSUPP:
      return @"socket does not support listen";
    case EINTR:
      return @"interrupted by signal";
    case EMFILE:
      return @"descriptor table is full";
    case EPROTONOSUPPORT:
      return @"The protocol is not supported by the address family or "
             @"implementation";
    case EPROTOTYPE:
      return @"The socket type is not supported by the protocol";

    default:
      return [NSString stringWithCString:strerror(errorCode)];
  }
}
#endif

- (BOOL)listenWithBacklog:(int)_backlogSize {
  // throws
  //   NGSocketIsAlreadyListeningException  when the socket is in the listen state
  //   NGCouldNotListenException            when the listen call failed
  
  if ([self isListening]) {
    [[[NGSocketIsAlreadyListeningException alloc]
              initWithReason:@"already called listen" socket:self] raise];
    return NO;
  }

  if (listen([self fileDescriptor], _backlogSize) != 0) {
    NSString *reason;
    reason = [self reasonForLastError];
    reason = [@"Could not listen: %@" stringByAppendingString:reason];
    
    [[[NGCouldNotListenException alloc]
              initWithReason:reason socket:self] raise];
    return NO;
  }

  /* set backlog size (and mark socket as 'listening') */
  self->backlogSize = _backlogSize;
  return YES;
}

- (id<NGActiveSocket>)accept {
  // throws
  //   NGCouldNotAcceptException  when the socket is not listening
  //   NGCouldNotAcceptException  when the accept call failed

  id<NGActiveSocket> socket;
  *(&socket) = nil;
  
  if (![self isListening]) {
    [[[NGCouldNotAcceptException alloc]
              initWithReason:@"socket is not listening" socket:self] raise];
  }
  
  SYNCHRONIZED(self->acceptLock) {
    id<NGSocketAddress> local  = nil;
    id<NGSocketAddress> remote = nil;
    int  len;
    char *data;
    int  newFd = NGInvalidSocketDescriptor;

    len   = [[self domain] addressRepresentationSize];
    data = calloc(1, len + 1);
    
    if ((newFd = accept(fd, (void *)data, &len)) == -1) {
      // call failed
      NSString *reason = nil;
      reason = [self reasonForLastError];
      reason = [@"Could not accept: " stringByAppendingString:reason];

      [[[NGCouldNotAcceptException alloc]
                initWithReason:reason socket:self] raise];
    }

    /* produce remote socket address object */
    remote = [[self domain] addressWithRepresentation:(void *)data
                            size:len];
    
    // getsockname if wildcard-IP-bind to get local IP address assigned
    // to the connection
    len = [[self domain] addressRepresentationSize];
    if (getsockname(newFd, (void *)data, &len) != 0) { // function is MT-safe
      [[[NGSocketException alloc]
                initWithReason:@"could not get local socket name" socket:self]
                raise];
    }
    local = [[self domain] addressWithRepresentation:(void *)data size:len];

    if (data) {
      free(data);
      data = NULL;
    }
    
    socket = [[NGActiveSocket allocWithZone:[self zone]]
                              _initWithDescriptor:newFd
                              localAddress:local
                              remoteAddress:remote];
    AUTORELEASE(socket);
  }
  END_SYNCHRONIZED;
  return socket;
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:@"<PassiveSocket: address=%@>",
                     [self localAddress]];
}

@end /* NGPassiveSocket */
