/* 
   NGActiveSocket.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGActiveSocket.h,v 1.2 2002/12/27 09:45:33 helge Exp $

#ifndef __NGNet_NGActiveSocket_H__
#define __NGNet_NGActiveSocket_H__

#import <Foundation/NSDate.h>
#include <NGStreams/NGSocket.h>
#include <NGStreams/NGSocketProtocols.h>
#include <NGStreams/NGStreamProtocols.h>

@class NSData, NSFileHandle;

/*
  Represents an active STREAM socket based on the standard Unix sockets library.

  An active socket can be either a socket gained by calling accept with an
  passive socket or by explicitly connecting one to an address (a client socket).
  Therefore an active socket has two addresses, the local and the remote one.

  There are three methods to perform a close, this is rooted in the fact that
  a socket actually is full-duplex, it provides a send and a receive channel.
  The stream-mode is updated according to what channels are open/closed.
  Initially the socket is full-duplex and you cannot reopen a channel that was
  shutdown. If you have shutdown both channels the socket can be considered
  closed.
*/

@interface NGActiveSocket : NGSocket < NGActiveSocket >
{
@private
  id<NGSocketAddress> remoteAddress;
  NGStreamMode        mode;
  
  NSTimeInterval receiveTimeout;
  NSTimeInterval sendTimeout;
}

+ (id)socketConnectedToAddress:(id<NGSocketAddress>)_address;
- (id)initWithDomain:(id<NGSocketDomain>)_domain; // designated initializer

#if !defined(WIN32)
+ (BOOL)socketPair:(id<NGSocket>[2])_pair inDomain:(id<NGSocketDomain>)_domain;
#endif

// ******************** operations ********************

// throws
//   NGSocketAlreadyConnectedException  when the socket is already connected
//   NGInvalidSocketDomainException     when the remote domain != local domain
//   NGCouldNotCreateSocketException    if the socket creation failed
- (BOOL)connectToAddress:(id<NGSocketAddress>)_address;

- (BOOL)shutdown; // do a complete shutdown
- (BOOL)shutdownSendChannel;
- (BOOL)shutdownReceiveChannel;

// ******************** accessors *********************

- (id<NGSocketAddress>)remoteAddress;
- (BOOL)isConnected;
- (BOOL)isOpen;

- (void)setSendTimeout:(NSTimeInterval)_timeout;
- (NSTimeInterval)sendTimeout;
- (void)setReceiveTimeout:(NSTimeInterval)_timeout;
- (NSTimeInterval)receiveTimeout;

// test whether a read, a write or both would block the thread (using select)
- (BOOL)wouldBlockInMode:(NGStreamMode)_mode;
- (int)waitForMode:(NGStreamMode)_mode timeout:(NSTimeInterval)_timeout;
- (unsigned)numberOfAvailableBytesForReading;
- (BOOL)isAlive;

// ******************** NGStream **********************

// throws
//   NGStreamReadErrorException    when the read call failed
//   NGSocketNotConnectedException when the socket is not connected
//   NGEndOfStreamException        when the end of the stream is reached
- (unsigned)readBytes:(void *)_buf count:(unsigned)_len;

// throws
//   NGStreamWriteErrorException   when the write call failed
//   NGSocketNotConnectedException when the socket is not connected
- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len;

- (BOOL)flush;                  // does nothing, sockets are unbuffered
- (NGStreamMode)mode;           // returns read/write

- (BOOL)safeReadBytes:(void *)_buf count:(unsigned)_len;
- (BOOL)safeWriteBytes:(const void *)_buf count:(unsigned)_len;

@end

@interface NGActiveSocket(DataMethods)

- (NSData *)readDataOfLength:(unsigned int)_length;
- (NSData *)safeReadDataOfLength:(unsigned int)_length;
- (unsigned int)writeData:(NSData *)_data;
- (BOOL)safeWriteData:(NSData *)_data;

@end

#endif /* __NGNet_NGActiveSocket_H__ */
