/* 
   NGStringTextStream.m
   
   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStringTextStream.m,v 1.3 2003/01/20 10:42:36 helge Exp $

#include "common.h"
#include "NGStringTextStream.h"
#include "NGStreamExceptions.h"

@implementation NGStringTextStream

+ (id)textStreamWithString:(NSString *)_string {
  return [[[self alloc] initWithString:_string] autorelease];
}
- (id)initWithString:(NSString *)_string {
  if ((self = [super init])) {
    self->string    = [_string retain];
    self->index     = 0;
    self->isMutable = [_string isKindOfClass:[NSMutableString class]];
  }
  return self;
}

- (void)dealloc {
  [self->string release];
  [super dealloc];
}

/* accessors */

- (NSString *)string {
  return string;
}

/* operations */

- (BOOL)close {
  // releases string
  [self->string release]; self->string = nil;
  return YES;
}

// NGTextInputStream

- (unichar)readCharacter {
  // throws
  //   NGStreamNotOpenException  when the stream is not open
  unsigned currentLength = [string length];
  unichar  result;

  if (string == nil) {
    [NGStreamNotOpenException raiseWithReason:
       @"tried to read from a string text stream which was closed"];
  }
  
  if (currentLength == index)
    [[[NGEndOfStreamException alloc] init] raise];

  result = [string characterAtIndex:index];
  index++;
  return result;
}

// NGExtendedTextInputStream

- (NSString *)readLineAsString {
  // throws
  //   NGStreamNotOpenException  when the stream is not open
  
  unsigned currentLength = [string length];
  NSRange  range;
  
  if (string == nil) {
    [NGStreamNotOpenException raiseWithReason:
            @"tried to read from a string text stream which was closed"];
  }
  
  if (currentLength == index)
    //[[[NGEndOfStreamException alloc] init] raise]
    return nil;
  
  range.location = index;
  range.length   = (currentLength - index);

  range = [string rangeOfString:@"\n" options:NSLiteralSearch range:range];
  if (range.length == 0) { // did not found newline
    NSString *result = [string substringFromIndex:index];
    index = currentLength;
    return result;
  }
  else {
    NSString *result = nil;

    range.length   = (range.location - index);
    range.location = index;

    result = [string substringWithRange:range];

    index += range.length + 1;

    return result;
  }
}

// NGTextOutputStream

- (BOOL)writeCharacter:(unichar)_character {
  // throws
  //   NGReadOnlyStreamException when the stream is not writeable
  //   NGStreamNotOpenException  when the stream is not open
  
  if (string == nil) {
    [NGStreamNotOpenException raiseWithReason:
            @"tried to write to a string text stream which was closed"];
    return NO;
  }
  if (!isMutable) {
    [[[NGReadOnlyStreamException alloc] init] raise];
    return NO;
  }
  
  [(NSMutableString *)string appendString:
    [NSString stringWithCharacters:&_character length:1]];
  return YES;
}

- (BOOL)writeString:(NSString *)_string {
  // throws
  //   NGReadOnlyStreamException when the stream is not writeable
  //   NGStreamNotOpenException  when the stream is not open
  
  if (string == nil) {
    [NGStreamNotOpenException raiseWithReason:
            @"tried to write to a string text stream which was closed"];
    return NO;
  }
  if (!isMutable) {
    [[[NGReadOnlyStreamException alloc] init] raise];
    return NO;
  }
  
  [(NSMutableString *)string appendString:_string];
  return YES;
}

- (BOOL)flush {
  return YES;
}

@end /* NGStringTextStream */
