/* 
   NGTextStream.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGTextStream.m,v 1.3 2002/12/27 09:49:00 helge Exp $

#include "common.h"
#include "NGTextStream.h"

@implementation NGTextStream

+ (int)version {
  return 2;
}

- (void)dealloc {
  RELEASE(self->lastException);
  [super dealloc];
}

/* NGTextInputStream */

- (NSException *)lastException {
  return nil;
}
- (void)setLastException:(NSException *)_exception {
  ASSIGN(self->lastException, _exception);
}
- (void)resetLastException {
  ASSIGN(self->lastException, (id)nil);
}

- (unichar)readCharacter {
  [self subclassResponsibility:_cmd];
  return 0;
}

- (BOOL)isOpen {
  return YES;
}

// NGExtendedTextInputStream

- (NSString *)readLineAsString {
  NSMutableString *str;
  unichar c;

  *(&str) = (id)[NSMutableString string];

  NS_DURING {
    while ((c = [self readCharacter]) != '\n')
      [str appendString:[NSString stringWithCharacters:&c length:1]];
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]]) {
      if ([str length] == 0) str = nil;
    }
    else
      [localException raise];
  }
  NS_ENDHANDLER;
  
  return str;
}

- (unsigned)readCharacters:(unichar *)_chars count:(unsigned)_count {
  /*
    Read up to _count characters, but one at the minimum.
  */
  volatile unsigned pos;

  NS_DURING {
    for (pos = 0; pos < _count; pos++)
      _chars[pos] = [self readCharacter];
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]]) {
      if (pos == 0)
        [localException raise];
    }
    else
      [localException raise];
  }
  NS_ENDHANDLER;

  NSAssert1(pos > 0, @"invalid character count to be returned: %i", pos);
  return pos;
}

- (BOOL)safeReadCharacters:(unichar *)_chars count:(unsigned)_count {
  volatile unsigned pos;
  
  for (pos = 0; pos < _count; pos++)
    _chars[pos] = [self readCharacter];
  
  return YES;
}

// NGTextOutputStream

- (BOOL)writeCharacter:(unichar)_character {
  [self subclassResponsibility:_cmd];
  return NO;
}

- (BOOL)writeString:(NSString *)_string {
  unsigned length = [_string length], cnt = 0;
  unichar  buffer[length];
  void     (*writeChar)(id, SEL, unichar);

  writeChar = (void (*)(id,SEL,unichar))
    [self methodForSelector:@selector(writeCharacter:)];
  
  [_string getCharacters:buffer];
  for (cnt = 0; cnt < length; cnt++)
    writeChar(self, @selector(writeCharacter:), buffer[cnt]);
  
  return YES;
}

- (BOOL)flush {
  return YES;
}

// NGExtendedTextOutputStream

- (BOOL)writeFormat:(NSString *)_format arguments:(va_list)_ap {
  NSString *tmp = [[NSString alloc] initWithFormat:_format arguments:_ap];
  tmp = AUTORELEASE(tmp);
  [self writeString:tmp];
  return YES;
}
- (BOOL)writeFormat:(NSString *)_format, ... {
  va_list ap;
  BOOL res = NO;

  va_start(ap, _format);

  NS_DURING {
    res = [self writeFormat:_format arguments:ap];
  }
  NS_HANDLER {
    va_end(ap);
    [localException raise];
  }
  NS_ENDHANDLER;
  va_end(ap);

  return res;
}

- (BOOL)writeNewline {
  if (![self writeString:@"\n"]) return NO;
  return [self flush];
}

- (unsigned)writeCharacters:(const unichar *)_chars count:(unsigned)_count {
  /*
    Write up to _count characters, but one at the minimum.
  */
  volatile unsigned pos;

  NS_DURING {
    for (pos = 0; pos < _count; pos++)
      [self writeCharacter:_chars[pos]];
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]]) {
      if (pos == 0)
        [localException raise];
    }
    else
      [localException raise];
  }
  NS_ENDHANDLER;

  NSAssert1(pos > 0, @"invalid character count to be returned: %i", pos);
  return pos;
}

- (BOOL)safeWriteCharacters:(const unichar *)_chars count:(unsigned)_count {
  unsigned pos;

  for (pos = 0; pos < _count; pos++) {
    if (![self writeCharacter:_chars[pos]])
      return NO;
  }
  
  return YES;
}

@end /* NGTextStream */
