// $Id: iCalDataSource.m,v 1.1 2003/02/12 17:23:27 helge Exp $

#include "iCalDataSource.h"
#include "iCalObject.h"
#include "iCalCalendar.h"
#include <SaxObjC/SaxObjC.h>
#include "common.h"

@interface NSObject(suppressCapitalizedKeyWarning)
+ (void)suppressCapitalizedKeyWarning;
@end

@implementation iCalDataSource

// THREAD
static id<NSObject,SaxXMLReader> parser = nil;
static SaxObjectDecoder *sax = nil;

- (void)_setupGlobals {
  if (parser == nil ) {
    SaxXMLReaderFactory *factory = 
      [SaxXMLReaderFactory standardXMLReaderFactory];
    parser = [[factory createXMLReaderForMimeType:@"text/calendar"] retain];
  }
  if (sax == nil && parser != nil) {
    NSBundle *bundle;
    NSString *p;
    
#if COCOA_Foundation_LIBRARY
    /* otherwise we get warning on "X-WR-TIMEZONE" etc */
    [NSClassFromString(@"NSKeyBinding") suppressCapitalizedKeyWarning];
#endif
    
    bundle = [NSBundle bundleForClass:[self class]];
    if ((p = [bundle pathForResource:@"NGiCal" ofType:@"xmap"]))
      sax = [[SaxObjectDecoder alloc] initWithMappingAtPath:p];
    else
      sax = [[SaxObjectDecoder alloc] initWithMappingNamed:@"NGiCal"];
    
    [parser setContentHandler:sax];
    [parser setErrorHandler:sax];
  }
}


- (id)initWithURL:(NSURL *)_url entityName:(NSString *)_ename {
  [self _setupGlobals];
  if ((self = [super init])) {
    self->url        = [_url copy];
    self->entityName = [_ename copy];
  }
  return self;
}
- (id)initWithPath:(NSString *)_path entityName:(NSString *)_ename {
  NSURL *lurl;
  
  lurl = [[[NSURL alloc] initFileURLWithPath:_path] autorelease];
  return [self initWithURL:lurl entityName:_ename];
}

- (id)initWithURL:(NSURL *)_url {
  return [self initWithURL:_url entityName:nil];
}
- (id)initWithPath:(NSString *)_path {
  return [self initWithPath:_path entityName:nil];
}
- (id)init {
  return [self initWithURL:nil];
}

- (void)dealloc {
  [self->fetchSpecification release];
  [self->entityName release];
  [self->url release];
  [super dealloc];
}

/* accessors */

- (void)setFetchSpecification:(EOFetchSpecification *)_fspec {
  if ([self->fetchSpecification isEqual:_fspec])
    return;
  
  ASSIGNCOPY(self->fetchSpecification, _fspec);

  [self postDataSourceChangedNotification];
}
- (EOFetchSpecification *)fetchSpecification {
  return self->fetchSpecification;
}

/* fetching */

- (id)_parseCalendar {
  id cal;
  
  if ((parser == nil) || (sax == nil)) {
    [self logWithFormat:@"no iCal parser or no SAX handler is available !"];
    return nil;
  }
  
  [parser parseFromSource:self->url];
  cal = [sax rootObject];
  
  return cal;
}

- (NSArray *)objectsFromCalendar:(id)_cal {
  NSString *ename;

  ename = [self->fetchSpecification entityName];
  if ([ename length] == 0)
    ename = self->entityName;

  if ([ename length] == 0)
    return [_cal allObjects];
  
  if ([ename isEqualToString:@"vevent"])
    return [_cal events];
  if ([ename isEqualToString:@"vtodo"])
    return [_cal todos];
  if ([ename isEqualToString:@"vjournal"])
    return [_cal journals];
  if ([ename isEqualToString:@"vfreebusy"])
    return [_cal freeBusys];
  
  [self logWithFormat:@"unknown calendar entity '%@'", ename];
  
  return nil;
}

- (NSArray *)fetchObjects {
  NSAutoreleasePool *pool;
  NSArray *result;
  id calendar;

  pool = [[NSAutoreleasePool alloc] init];
  
  if ((calendar = [self _parseCalendar]) == nil)
    return nil;
  
  if (self->fetchSpecification == nil) {
    result = [[self objectsFromCalendar:calendar] shallowCopy];
  }
  else {
    NSMutableArray *ma;
    NSEnumerator   *e;
    EOQualifier *q;
    NSArray     *sort;
    NSArray     *objects;
    iCalObject  *object;
    
    /* get objects */
    
    objects = [self objectsFromCalendar:calendar];
    
    /* first filter using qualifier */
    
    ma = [NSMutableArray arrayWithCapacity:[objects count]];
    q  = [self->fetchSpecification qualifier];
    e  = [objects objectEnumerator];
    while ((object = [e nextObject])) {
      if (q) {
	if (![(id<EOQualifierEvaluation>)q evaluateWithObject:object])
	  continue;
      }
      
      [ma addObject:object];
    }
    
    /* now sort */
    
    if ((sort = [self->fetchSpecification sortOrderings]))
      [ma sortUsingKeyOrderArray:sort];

    result = [ma shallowCopy];
  }
  
  [pool release];
  
  return [result autorelease];
}

@end /* iCalDataSource */
